"""
Base typeclass for in-game Channels.

"""
import re

from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils.text import slugify

import evennia
from evennia.comms.managers import ChannelManager
from evennia.comms.models import ChannelDB
from evennia.typeclasses.models import TypeclassBase
from evennia.utils import create, logger
from evennia.utils.utils import inherits_from, make_iter


class DefaultChannel(ChannelDB, metaclass=TypeclassBase):
    """
    This is the base class for all Channel Comms. Inherit from this to
    create different types of communication channels.

    Class-level variables:
    - `send_to_online_only` (bool, default True) - if set, will only try to
      send to subscribers that are actually active. This is a useful optimization.
    - `log_file` (str, default `"channel_{channelname}.log"`). This is the
      log file to which the channel history will be saved. The `{channelname}` tag
      will be replaced by the key of the Channel. If an Attribute 'log_file'
      is set, this will be used instead. If this is None and no Attribute is found,
      no history will be saved.
    - `channel_prefix_string` (str, default `"[{channelname} ]"`) - this is used
      as a simple template to get the channel prefix with `.channel_prefix()`. It is used
      in front of every channel message; use `{channelmessage}` token to insert the
      name of the current channel. Set to `None` if you want no prefix (or want to
      handle it in a hook during message generation instead.
    - `channel_msg_nick_pattern`(str, default `"{alias}\\s*?|{alias}\\s+?(?P<arg1>.+?)") -
      this is what used when a channel subscriber gets a channel nick assigned to this
      channel. The nickhandler uses the pattern to pick out this channel's name from user
      input. The `{alias}` token will get both the channel's key and any set/custom aliases
      per subscriber. You need to allow for an `<arg1>` regex group to catch any message
      that should be send to the  channel. You usually don't need to change this pattern
      unless you are changing channel command-style entirely.
    - `channel_msg_nick_replacement` (str, default `"channel {channelname} = $1"` - this
      is used by the nickhandler to generate a replacement string once the nickhandler (using
      the `channel_msg_nick_pattern`) identifies that the channel should be addressed
      to send a message to it. The `<arg1>` regex pattern match from `channel_msg_nick_pattern`
      will end up at the `$1` position in the replacement. Together, this allows you do e.g.
      'public Hello' and have that become a mapping to `channel public = Hello`. By default,
      the account-level `channel` command is used. If you were to rename that command you must
      tweak the output to something like `yourchannelcommandname {channelname} = $1`.

    """

    objects = ChannelManager()

    # channel configuration

    # only send to characters/accounts who has an active session (this is a
    # good optimization since people can still recover history separately).
    send_to_online_only = True
    # store log in log file. `channel_key tag will be replace with key of channel.
    # Will use log_file Attribute first, if given
    log_file = "channel_{channelname}.log"
    # which prefix to use when showing were a message is coming from. Set to
    # None to disable and set this later.
    channel_prefix_string = "[{channelname}] "

    # default nick-alias replacements (default using the 'channel' command)
    channel_msg_nick_pattern = r"{alias}\s*?|{alias}\s+?(?P<arg1>.+?)"
    channel_msg_nick_replacement = "@channel {channelname} = $1"

    def at_first_save(self):
        """
        Called by the typeclass system the very first time the channel
        is saved to the database. Generally, don't overload this but
        the hooks called by this method.

        """
        self.basetype_setup()
        self.at_channel_creation()
        # initialize Attribute/TagProperties
        self.init_evennia_properties()

        if hasattr(self, "_createdict"):
            # this is only set if the channel was created
            # with the utils.create.create_channel function.
            cdict = self._createdict
            if not cdict.get("key"):
                if not self.db_key:
                    self.db_key = "#i" % self.dbid
            elif cdict["key"] and self.key != cdict["key"]:
                self.key = cdict["key"]
            if cdict.get("aliases"):
                self.aliases.add(cdict["aliases"])
            if cdict.get("locks"):
                self.locks.add(cdict["locks"])
            if cdict.get("keep_log"):
                self.attributes.add("keep_log", cdict["keep_log"])
            if cdict.get("desc"):
                self.attributes.add("desc", cdict["desc"])
            if cdict.get("tags"):
                self.tags.batch_add(*cdict["tags"])
            if cdict.get("attrs"):
                self.attributes.batch_add(*cdict["attrs"])

    def basetype_setup(self):
        self.locks.add("send:all();listen:all();control:perm(Admin)")

        # make sure we don't have access to a same-named old channel's history.
        log_file = self.get_log_filename()
        logger.rotate_log_file(log_file, num_lines_to_append=0)

    def at_channel_creation(self):
        """
        Called once, when the channel is first created.

        """
        pass

    # helper methods, for easy overloading

    _log_file = None

    def get_log_filename(self):
        """
        File name to use for channel log.

        Returns:
            str: The filename to use (this is always assumed to be inside
                settings.LOG_DIR)

        """
        if not self._log_file:
            self._log_file = self.attributes.get(
                "log_file", self.log_file.format(channelname=self.key.lower())
            )
        return self._log_file

    def set_log_filename(self, filename):
        """
        Set a custom log filename.

        Args:
            filename (str): The filename to set. This is a path starting from
                inside the settings.LOG_DIR location.

        """
        self.attributes.add("log_file", filename)

    def has_connection(self, subscriber):
        """
        Checks so this account is actually listening
        to this channel.

        Args:
            subscriber (Account or Object): Entity to check.

        Returns:
            has_sub (bool): Whether the subscriber is subscribing to
                this channel or not.

        Notes:
            This will first try Account subscribers and only try Object
                if the Account fails.

        """
        has_sub = self.subscriptions.has(subscriber)
        if not has_sub and inherits_from(subscriber, evennia.DefaultObject):
            # it's common to send an Object when we
            # by default only allow Accounts to subscribe.
            has_sub = self.subscriptions.has(subscriber.account)
        return has_sub

    @property
    def mutelist(self):
        return self.db.mute_list or []

    @property
    def banlist(self):
        return self.db.ban_list or []

    @property
    def wholist(self):
        subs = self.subscriptions.all()
        muted = list(self.mutelist)
        listening = [ob for ob in subs if ob.is_connected and ob not in muted]
        if subs:
            # display listening subscribers in bold
            string = ", ".join(
                [
                    account.key if account not in listening else f"|w{account.key}|n"
                    for account in subs
                ]
            )
        else:
            string = "<None>"
        return string

    def mute(self, subscriber, **kwargs):
        """
        Adds an entity to the list of muted subscribers.
        A muted subscriber will no longer see channel messages,
        but may use channel commands.

        Args:
            subscriber (Object or Account): Subscriber to mute.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            bool: True if muting was successful, False if we were already
                muted.

        """
        mutelist = self.mutelist
        if subscriber not in mutelist:
            mutelist.append(subscriber)
            self.db.mute_list = mutelist
            return True
        return False

    def unmute(self, subscriber, **kwargs):
        """
        Removes an entity from the list of muted subscribers.  A muted subscriber
        will no longer see channel messages, but may use channel commands.

        Args:
            subscriber (Object or Account): The subscriber to unmute.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            bool: True if unmuting was successful, False if we were already
                unmuted.

        """
        mutelist = self.mutelist
        if subscriber in mutelist:
            mutelist.remove(subscriber)
            return True
        return False

    def ban(self, target, **kwargs):
        """
        Ban a given user from connecting to the channel. This will not stop
        users already connected, so the user must be booted for this to take
        effect.

        Args:
            target (Object or Account): The entity to unmute. This need not
                be a subscriber.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            bool: True if banning was successful, False if target was already
                banned.
        """
        banlist = self.banlist
        if target not in banlist:
            banlist.append(target)
            self.db.ban_list = banlist
            return True
        return False

    def unban(self, target, **kwargs):
        """
        Un-Ban a given user. This will not reconnect them - they will still
        have to reconnect and set up aliases anew.

        Args:
            target (Object or Account): The entity to unmute. This need not
                be a subscriber.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            bool: True if unbanning was successful, False if target was not
                previously banned.
        """
        banlist = list(self.banlist)
        if target in banlist:
            banlist = [banned for banned in banlist if banned != target]
            self.db.ban_list = banlist
            return True
        return False

    def connect(self, subscriber, **kwargs):
        """
        Connect the user to this channel. This checks access.

        Args:
            subscriber (Account or Object): the entity to subscribe
                to this channel.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            success (bool): Whether or not the addition was
                successful.

        """
        # check access
        if subscriber in self.banlist or not self.access(subscriber, "listen"):
            return False
        # pre-join hook
        connect = self.pre_join_channel(subscriber)
        if not connect:
            return False
        # subscribe
        self.subscriptions.add(subscriber)
        # unmute
        self.unmute(subscriber)
        # post-join hook
        self.post_join_channel(subscriber)
        return True

    def disconnect(self, subscriber, **kwargs):
        """
        Disconnect entity from this channel.

        Args:
            subscriber (Account of Object): the
                entity to disconnect.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            success (bool): Whether or not the removal was
                successful.

        """
        # pre-disconnect hook
        disconnect = self.pre_leave_channel(subscriber)
        if not disconnect:
            return False
        # disconnect
        self.subscriptions.remove(subscriber)
        # unmute
        self.unmute(subscriber)
        # post-disconnect hook
        self.post_leave_channel(subscriber)
        return True

    def access(
        self,
        accessing_obj,
        access_type="listen",
        default=False,
        no_superuser_bypass=False,
        **kwargs,
    ):
        """
        Determines if another object has permission to access.

        Args:
            accessing_obj (Object): Object trying to access this one.
            access_type (str, optional): Type of access sought.
            default (bool, optional): What to return if no lock of access_type was found
            no_superuser_bypass (bool, optional): Turns off superuser
                lock bypass. Be careful with this one.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            return (bool): Result of lock check.

        """
        return self.locks.check(
            accessing_obj,
            access_type=access_type,
            default=default,
            no_superuser_bypass=no_superuser_bypass,
        )

    @classmethod
    def create(cls, key, creator=None, *args, **kwargs):
        """
        Creates a basic Channel with default parameters, unless otherwise
        specified or extended.

        Provides a friendlier interface to the utils.create_channel() function.

        Args:
            key (str): This must be unique.
            creator (Account or Object): Entity to associate with this channel
                (used for tracking)

        Keyword Args:
            aliases (list of str): List of alternative (likely shorter) keynames.
            description (str): A description of the channel, for use in listings.
            locks (str): Lockstring.
            keep_log (bool): Log channel throughput.
            typeclass (str or class): The typeclass of the Channel (not
                often used).
            ip (str): IP address of creator (for object auditing).

        Returns:
            channel (Channel): A newly created Channel.
            errors (list): A list of errors in string form, if any.

        """
        errors = []
        obj = None
        ip = kwargs.pop("ip", "")

        try:
            kwargs["desc"] = kwargs.pop("description", "")
            kwargs["typeclass"] = kwargs.get("typeclass", cls)
            obj = create.create_channel(key, *args, **kwargs)

            # Record creator id and creation IP
            if ip:
                obj.db.creator_ip = ip
            if creator:
                obj.db.creator_id = creator.id

        except Exception as exc:
            errors.append("An error occurred while creating this '%s' object." % key)
            logger.log_err(exc)

        return obj, errors

    def delete(self):
        """
        Deletes channel.

        Returns:
            bool: If deletion was successful. Only time it can fail would be
                if channel was already deleted. Even if it were to fail, all subscribers
                will be disconnected.

        """
        self.attributes.clear()
        self.aliases.clear()
        for subscriber in self.subscriptions.all():
            self.disconnect(subscriber)
        if not self.pk:
            return False
        super().delete()
        return True

    def channel_prefix(self):
        """
        Hook method. How the channel should prefix itself for users.

        Returns:
            str: The channel prefix.

        """
        return self.channel_prefix_string.format(channelname=self.key)

    def add_user_channel_alias(self, user, alias, **kwargs):
        """
        Add a personal user-alias for this channel to a given subscriber.

        Args:
            user (Object or Account): The one to alias this channel.
            alias (str): The desired alias.

        Note:
            This is tightly coupled to the default `channel` command. If you
            change that, you need to change this as well.

            We add two nicks - one is a plain `alias -> channel.key` that
            users need to be able to reference this channel easily. The other
            is a templated nick to easily be able to send messages to the
            channel without needing to give the full `channel` command. The
            structure of this nick is given by `self.channel_msg_nick_pattern`
            and `self.channel_msg_nick_replacement`. By default it maps
            `alias <msg> -> channel <channelname> = <msg>`, so that you can
            for example just write `pub Hello` to send a message.

            The alias created is `alias $1 -> channel channel = $1`, to allow
            for sending to channel using the main channel command.

        """
        chan_key = self.key.lower()

        # the message-pattern allows us to type the channel on its own without
        # needing to use the `channel` command explicitly.
        msg_nick_pattern = self.channel_msg_nick_pattern.format(alias=re.escape(alias))
        msg_nick_replacement = self.channel_msg_nick_replacement.format(channelname=chan_key)
        user.nicks.add(
            msg_nick_pattern,
            msg_nick_replacement,
            category="inputline",
            pattern_is_regex=True,
            **kwargs,
        )

        if chan_key != alias:
            # this allows for using the alias for general channel lookups
            user.nicks.add(alias, chan_key, category="channel", **kwargs)

    @classmethod
    def remove_user_channel_alias(cls, user, alias, **kwargs):
        """
        Remove a personal channel alias from a user.

        Args:
           user (Object or Account): The user to remove an alias from.
           alias (str): The alias to remove.
           **kwargs: Unused by default. Can be used to pass extra variables
                into a custom implementation.

        Notes:
            The channel-alias actually consists of two aliases - one
            channel-based one for searching channels with the alias and one
            inputline one for doing the 'channelalias msg' - call.

            This is a classmethod because it doesn't actually operate on the
            channel instance.

            It sits on the channel because the nick structure for this is
            pretty complex and needs to be located in a central place (rather
            on, say, the channel command).

        """
        user.nicks.remove(alias, category="channel", **kwargs)
        msg_nick_pattern = cls.channel_msg_nick_pattern.format(alias=alias)
        user.nicks.remove(msg_nick_pattern, category="inputline", **kwargs)

    def at_pre_msg(self, message, **kwargs):
        """
        Called before the starting of sending the message to a receiver. This
        is called before any hooks on the receiver itself. If this returns
        None/False, the sending will be aborted.

        Args:
            message (str): The message to send.
            **kwargs (any): Keywords passed on from `.msg`. This includes
                `senders`.

        Returns:
            str, False or None: Any custom changes made to the message. If
                falsy, no message will be sent.

        """
        return message

    def msg(self, message, senders=None, bypass_mute=False, **kwargs):
        """
        Send message to channel, causing it to be distributed to all non-muted
        subscribed users of that channel.

        Args:
            message (str): The message to send.
            senders (Object, Account or list, optional): If not given, there is
                no way to associate one or more senders with the message (like
                a broadcast message or similar).
            bypass_mute (bool, optional): If set, always send, regardless of
                individual mute-state of subscriber. This can be used for
                global announcements or warnings/alerts.
            **kwargs (any): This will be passed on to all hooks. Use `no_prefix`
                to exclude the channel prefix.

        Notes:
            The call hook calling sequence is:

            - `msg = channel.at_pre_msg(message, **kwargs)` (aborts for all if return None)
            - `msg = receiver.at_pre_channel_msg(msg, channel, **kwargs)` (aborts for receiver if return None)
            - `receiver.at_channel_msg(msg, channel, **kwargs)`
            - `receiver.at_post_channel_msg(msg, channel, **kwargs)``
            Called after all receivers are processed:
            - `channel.at_post_all_msg(message, **kwargs)`

            (where the senders/bypass_mute are embedded into **kwargs for
            later access in hooks)

        """
        senders = make_iter(senders) if senders else []
        if self.send_to_online_only:
            receivers = self.subscriptions.online()
        else:
            receivers = self.subscriptions.all()
        if not bypass_mute:
            receivers = [receiver for receiver in receivers if receiver not in self.mutelist]

        send_kwargs = {"senders": senders, "bypass_mute": bypass_mute, **kwargs}

        # pre-send hook
        message = self.at_pre_msg(message, **send_kwargs)
        if message in (None, False):
            return

        for receiver in receivers:
            # send to each individual subscriber

            try:
                recv_message = receiver.at_pre_channel_msg(message, self, **send_kwargs)
                if recv_message in (None, False):
                    return

                receiver.channel_msg(recv_message, self, **send_kwargs)

                receiver.at_post_channel_msg(recv_message, self, **send_kwargs)

            except Exception:
                logger.log_trace(f"Error sending channel message to {receiver}.")

        # post-send hook
        self.at_post_msg(message, **send_kwargs)

    def at_post_msg(self, message, **kwargs):
        """
        This is called after sending to *all* valid recipients. It is normally
        used for logging/channel history.

        Args:
            message (str): The message sent.
            **kwargs (any): Keywords passed on from `msg`, including `senders`.

        """
        # save channel history to log file
        log_file = self.get_log_filename()
        if log_file:
            senders = ",".join(sender.key for sender in kwargs.get("senders", []))
            senders = f"{senders}: " if senders else ""
            message = f"{senders}{message}"
            logger.log_file(message, log_file)

    def pre_join_channel(self, joiner, **kwargs):
        """
        Hook method. Runs right before a channel is joined. If this
        returns a false value, channel joining is aborted.

        Args:
            joiner (object): The joining object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            should_join (bool): If `False`, channel joining is aborted.

        """
        return True

    def post_join_channel(self, joiner, **kwargs):
        """
        Hook method. Runs right after an object or account joins a channel.

        Args:
            joiner (object): The joining object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Notes:
            By default this adds the needed channel nicks to the joiner.

        """
        key_and_aliases = [self.key.lower()] + [alias.lower() for alias in self.aliases.all()]
        for key_or_alias in key_and_aliases:
            self.add_user_channel_alias(joiner, key_or_alias, **kwargs)

    def pre_leave_channel(self, leaver, **kwargs):
        """
        Hook method. Runs right before a user leaves a channel. If this returns a false
        value, leaving the channel will be aborted.

        Args:
            leaver (object): The leaving object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        Returns:
            should_leave (bool): If `False`, channel parting is aborted.

        """
        return True

    def post_leave_channel(self, leaver, **kwargs):
        """
        Hook method. Runs right after an object or account leaves a channel.

        Args:
            leaver (object): The leaving object.
            **kwargs (dict): Arbitrary, optional arguments for users
                overriding the call (unused by default).

        """
        chan_key = self.key.lower()
        key_or_aliases = [self.key.lower()] + [alias.lower() for alias in self.aliases.all()]
        nicktuples = leaver.nicks.get(category="channel", return_tuple=True, return_list=True)
        key_or_aliases += [tup[2] for tup in nicktuples if tup[3].lower() == chan_key]
        for key_or_alias in key_or_aliases:
            self.remove_user_channel_alias(leaver, key_or_alias, **kwargs)

    def at_init(self):
        """
        Hook method. This is always called whenever this channel is
        initiated -- that is, whenever it its typeclass is cached from
        memory. This happens on-demand first time the channel is used
        or activated in some way after being created but also after
        each server restart or reload.

        """
        pass

    #
    # Web/Django methods
    #

    def web_get_admin_url(self):
        """
        Returns the URI path for the Django Admin page for this object.

        ex. Account#1 = '/admin/accounts/accountdb/1/change/'

        Returns:
            path (str): URI path to Django Admin page for object.

        """
        content_type = ContentType.objects.get_for_model(self.__class__)
        return reverse(
            "admin:%s_%s_change" % (content_type.app_label, content_type.model), args=(self.id,)
        )

    @classmethod
    def web_get_create_url(cls):
        """
        Returns the URI path for a View that allows users to create new
        instances of this object.

        ex. Chargen = '/characters/create/'

        For this to work, the developer must have defined a named view somewhere
        in urls.py that follows the format 'modelname-action', so in this case
        a named view of 'channel-create' would be referenced by this method.

        ex.
        url(r'channels/create/', ChannelCreateView.as_view(), name='channel-create')

        If no View has been created and defined in urls.py, returns an
        HTML anchor.

        This method is naive and simply returns a path. Securing access to
        the actual view and limiting who can create new objects is the
        developer's responsibility.

        Returns:
            path (str): URI path to object creation page, if defined.

        """
        try:
            return reverse("%s-create" % slugify(cls._meta.verbose_name))
        except Exception:
            return "#"

    def web_get_detail_url(self):
        """
        Returns the URI path for a View that allows users to view details for
        this object.

        ex. Oscar (Character) = '/characters/oscar/1/'

        For this to work, the developer must have defined a named view somewhere
        in urls.py that follows the format 'modelname-action', so in this case
        a named view of 'channel-detail' would be referenced by this method.

        ex.
        ::

            url(r'channels/(?P<slug>[\w\d\-]+)/$',
                ChannelDetailView.as_view(), name='channel-detail')

        If no View has been created and defined in urls.py, returns an
        HTML anchor.

        This method is naive and simply returns a path. Securing access to
        the actual view and limiting who can view this object is the developer's
        responsibility.

        Returns:
            path (str): URI path to object detail page, if defined.

        """
        try:
            return reverse(
                "%s-detail" % slugify(self._meta.verbose_name),
                kwargs={"slug": slugify(self.db_key)},
            )
        except Exception:
            return "#"

    def web_get_update_url(self):
        """
        Returns the URI path for a View that allows users to update this
        object.

        ex. Oscar (Character) = '/characters/oscar/1/change/'

        For this to work, the developer must have defined a named view somewhere
        in urls.py that follows the format 'modelname-action', so in this case
        a named view of 'channel-update' would be referenced by this method.

        ex.
        ::

            url(r'channels/(?P<slug>[\w\d\-]+)/(?P<pk>[0-9]+)/change/$',
                ChannelUpdateView.as_view(), name='channel-update')

        If no View has been created and defined in urls.py, returns an
        HTML anchor.

        This method is naive and simply returns a path. Securing access to
        the actual view and limiting who can modify objects is the developer's
        responsibility.

        Returns:
            path (str): URI path to object update page, if defined.

        """
        try:
            return reverse(
                "%s-update" % slugify(self._meta.verbose_name),
                kwargs={"slug": slugify(self.db_key)},
            )
        except Exception:
            return "#"

    def web_get_delete_url(self):
        """
        Returns the URI path for a View that allows users to delete this object.

        ex. Oscar (Character) = '/characters/oscar/1/delete/'

        For this to work, the developer must have defined a named view somewhere
        in urls.py that follows the format 'modelname-action', so in this case
        a named view of 'channel-delete' would be referenced by this method.

        ex.
        url(r'channels/(?P<slug>[\w\d\-]+)/(?P<pk>[0-9]+)/delete/$',
            ChannelDeleteView.as_view(), name='channel-delete')

        If no View has been created and defined in urls.py, returns an
        HTML anchor.

        This method is naive and simply returns a path. Securing access to
        the actual view and limiting who can delete this object is the developer's
        responsibility.

        Returns:
            path (str): URI path to object deletion page, if defined.

        """
        try:
            return reverse(
                "%s-delete" % slugify(self._meta.verbose_name),
                kwargs={"slug": slugify(self.db_key)},
            )
        except Exception:
            return "#"

    # Used by Django Sites/Admin
    get_absolute_url = web_get_detail_url

    # TODO Evennia 1.0+ removed hooks. Remove in 1.1.
    def message_transform(self, *args, **kwargs):
        raise RuntimeError(
            "Channel.message_transform is no longer used in 1.0+. "
            "Use Account/Object.at_pre_channel_msg instead."
        )

    def distribute_message(self, msgobj, online=False, **kwargs):
        raise RuntimeError("Channel.distribute_message is no longer used in 1.0+.")

    def format_senders(self, senders=None, **kwargs):
        raise RuntimeError(
            "Channel.format_senders is no longer used in 1.0+. "
            "Use Account/Object.at_pre_channel_msg instead."
        )

    def pose_transform(self, msgobj, sender_string, **kwargs):
        raise RuntimeError(
            "Channel.pose_transform is no longer used in 1.0+. "
            "Use Account/Object.at_pre_channel_msg instead."
        )

    def format_external(self, msgobj, senders, emit=False, **kwargs):
        raise RuntimeError(
            "Channel.format_external is no longer used in 1.0+. "
            "Use Account/Object.at_pre_channel_msg instead."
        )

    def format_message(self, msgobj, emit=False, **kwargs):
        raise RuntimeError(
            "Channel.format_message is no longer used in 1.0+. "
            "Use Account/Object.at_pre_channel_msg instead."
        )

    def pre_send_message(self, msg, **kwargs):
        raise RuntimeError("Channel.pre_send_message was renamed to Channel.at_pre_msg.")

    def post_send_message(self, msg, **kwargs):
        raise RuntimeError("Channel.post_send_message was renamed to Channel.at_post_msg.")



"""
# Dice

Rolls dice for roleplaying, in-game gambling or GM:ing

Evennia contribution - Griatch 2012

This module implements a a dice-roller and a `dice`/`roll` command
to go with it. It uses standard RPG 'd'-syntax (e.g. 2d6 to roll two
six-sided die) and also supports modifiers such as 3d6 + 5.

    > roll 1d20 + 2

One can also specify a standard Python operator in order to specify
eventual target numbers and get results in a fair and guaranteed
unbiased way.  For example a GM could (using the dice command) from
the start define the roll as 2d6 < 8 to show that a roll below 8 is
required to succeed. The command will normally echo this result to all
parties (although it also has options for hidden and secret rolls).


# Installation:


Add the `CmdDice` command from this module to your character's cmdset
(and then restart the server):

```python
# in mygame/commands/default_cmdsets.py

# ...
from evennia.contrib.rpg import dice  <---

class CharacterCmdSet(default_cmds.CharacterCmdSet):
    # ...
    def at_cmdset_creation(self):
        # ...
        self.add(dice.CmdDice())  # <---

```

# Usage:

    roll 1d100 + 10

To roll dice in code, use the `roll` function from this module:

    from evennia.contrib.rpg import dice

    dice.roll("3d10 + 2")


If your system generates the dice dynamically you can also enter each part
of the roll separately:

    dice.roll(3, 10, ("+", 2))  # 3d10 + 2


"""
import re
from ast import literal_eval
from random import randint

from evennia import CmdSet, default_cmds
from evennia.utils.utils import simple_eval


def roll(
    dice,
    dicetype=6,
    modifier=None,
    conditional=None,
    return_tuple=False,
    max_dicenum=10,
    max_dicetype=1000,
):
    """
    This is a standard dice roller.

    Args:
     dice (int or str): If an `int`, this is the number of dice to roll, and `dicetype` is used
        to determine the type. If a `str`, it should be on the form `NdM` where `N` is the number
        of dice and `M` is the number of sides on each die. Also
        `NdM [modifier] [number] [conditional]` is understood, e.g. `1d6 + 3`
        or `2d10 / 2 > 10`.
     dicetype (int, optional): Number of sides of the dice to be rolled. Ignored if
        `dice` is a string.
     modifier (tuple, optional): A tuple `(operator, value)`, where operator is
        one of `"+"`, `"-"`, `"/"` or `"*"`. The result of the dice
        roll(s) will be modified by this value. Ignored if `dice` is a string.
     conditional (tuple, optional): A tuple `(conditional, value)`, where
        conditional is one of `"=="`,`"<"`,`">"`,`">="`,`"<=`" or "`!=`".
        Ignored if `dice` is a string.
     return_tuple (bool): Return a tuple with all individual roll
        results or not.
     max_dicenum (int): The max number of dice to allow to be rolled.
     max_dicetype (int): The max number of sides on the dice to roll.

    Returns:
        int, bool or tuple : By default, this is the result of the roll + modifiers.  If
        `conditional` is given, or `dice` is a string defining a conditional, then a True/False
        value is returned. Finally, if `return_tuple` is set, this is a tuple
        `(result, outcome, diff, rolls)`, where, `result` is the the normal result of the
        roll + modifiers,  `outcome` and `diff` are the boolean absolute difference between the roll
        and the `conditional` input; both will be will be `None` if `conditional` is not set.
        The `rolls` a tuple holding all the individual rolls (one or more depending on how many
        dice were rolled).

    Raises:
        TypeError if non-supported modifiers or conditionals are given.

    Notes:
        All input numbers are converted to integers.

    Examples:
        ::
            # string form
            print roll("3d6 + 2")
            10
            print roll("2d10 + 2 > 10")
            True
            print roll("2d20 - 2 >= 10")
            (8, False, 2, (4, 6)) # roll was 4 + 6 - 2 = 8

            # explicit arguments
            print roll(2, 6) # 2d6
            7
            print roll(1, 100, ('+', 5) # 1d100 + 5
            4
            print roll(1, 20, conditional=('<', 10) # let'say we roll 3
            True
            print roll(3, 10, return_tuple=True)
            (11, None, None, (2, 5, 4))
            print roll(2, 20, ('-', 2), conditional=('>=', 10), return_tuple=True)
            (8, False, 2, (4, 6)) # roll was 4 + 6 - 2 = 8

    """

    modifier_string = ""
    conditional_string = ""
    conditional_value = None
    if isinstance(dice, str) and "d" in dice.lower():
        # A string is given, parse it as NdM dice notation
        roll_string = dice.lower()

        # split to get the NdM syntax
        dicenum, rest = roll_string.split("d", 1)

        # parse packwards right-to-left
        if any(True for cond in ("==", "<", ">", "!=", "<=", ">=") if cond in rest):
            # split out any conditionals, like '< 12'
            rest, *conditionals = re.split(r"(==|<=|>=|<|>|!=)", rest, maxsplit=1)
            try:
                conditional_value = int(conditionals[1])
            except ValueError:
                raise TypeError(
                    f"Conditional '{conditionals[-1]}' was not recognized. Must be a number."
                )
            conditional_string = "".join(conditionals)

        if any(True for op in ("+", "-", "*", "/") if op in rest):
            # split out any modifiers, like '+ 2'
            rest, *modifiers = re.split(r"(\+|-|/|\*)", rest, maxsplit=1)
            modifier_string = "".join(modifiers)

        # whatever is left is the dice type
        dicetype = rest

    else:
        # an integer is given - explicit modifiers and conditionals as part of kwargs
        dicenum = int(dice)
        dicetype = int(dicetype)
        if modifier:
            modifier_string = "".join(str(part) for part in modifier)
        if conditional:
            conditional_value = int(conditional[1])
            conditional_string = "".join(str(part) for part in conditional)

    try:
        dicenum = int(dicenum)
        dicetype = int(dicetype)
    except Exception:
        raise TypeError(
            f"The number of dice and dice-size must both be numerical. Got '{dicenum}' "
            f"and '{dicetype}'."
        )
    if 0 < dicenum > max_dicenum:
        raise TypeError(f"Invalid number of dice rolled (must be between 1 and {max_dicenum}).")
    if 0 < dicetype > max_dicetype:
        raise TypeError(f"Invalid die-size used (must be between 1 and {max_dicetype} sides).")

    # roll all dice, remembering each roll
    rolls = tuple([randint(1, dicetype) for _ in range(dicenum)])
    result = sum(rolls)

    if modifier_string:
        result = simple_eval(f"{result} {modifier_string}")

    outcome, diff = None, None
    if conditional_string and conditional_value:
        outcome = simple_eval(f"{result} {conditional_string}")
        diff = abs(result - conditional_value)

    if return_tuple:
        return result, outcome, diff, rolls
    elif conditional or (conditional_string and conditional_value):
        return outcome  # True|False
    else:
        return result  # integer


# legacy alias
roll_dice = roll


RE_PARTS = re.compile(r"(d|\+|-|/|\*|<|>|<=|>=|!=|==)")
RE_MOD = re.compile(r"(\+|-|/|\*)")
RE_COND = re.compile(r"(<|>|<=|>=|!=|==)")


class CmdDice(default_cmds.MuxCommand):
    """
    roll dice

    Usage:
      dice[/switch] <nr>d<sides> [modifier] [success condition]

    Switch:
      hidden - tell the room the roll is being done, but don't show the result
      secret - don't inform the room about neither roll nor result

    Examples:
      dice 3d6 + 4
      dice 1d100 - 2 < 50

    This will roll the given number of dice with given sides and modifiers.
    So e.g. 2d6 + 3 means to 'roll a 6-sided die 2 times and add the result,
    then add 3 to the total'.
    Accepted modifiers are +, -, * and /.
    A success condition is given as normal Python conditionals
    (<,>,<=,>=,==,!=). So e.g. 2d6 + 3 > 10 means that the roll will succeed
    only if the final result is above 8. If a success condition is given, the
    outcome (pass/fail) will be echoed along with how much it succeeded/failed
    with. The hidden/secret switches will hide all or parts of the roll from
    everyone but the person rolling.
    """

    key = "dice"
    aliases = ["roll", "@dice"]
    locks = "cmd:all()"

    def func(self):
        """Mostly parsing for calling the dice roller function"""

        if not self.args:
            self.caller.msg("Usage: @dice <nr>d<sides> [modifier] [conditional]")
            return
        argstring = "".join(str(arg) for arg in self.args)

        parts = [part for part in RE_PARTS.split(self.args) if part]
        len_parts = len(parts)
        modifier = None
        conditional = None

        if len_parts < 3 or parts[1] != "d":
            self.caller.msg(
                "You must specify the die roll(s) as <nr>d<sides>."
                " For example, 2d6 means rolling a 6-sided die 2 times."
            )
            return

        # Limit the number of dice and sides a character can roll to prevent server slow down and crashes
        ndicelimit = 10000  # Maximum number of dice
        nsidelimit = 10000  # Maximum number of sides
        if int(parts[0]) > ndicelimit or int(parts[2]) > nsidelimit:
            self.caller.msg("The maximum roll allowed is %sd%s." % (ndicelimit, nsidelimit))
            return

        ndice, nsides = parts[0], parts[2]
        if len_parts == 3:
            # just something like 1d6
            pass
        elif len_parts == 5:
            # either e.g. 1d6 + 3  or something like 1d6 > 3
            if parts[3] in ("+", "-", "*", "/"):
                modifier = (parts[3], parts[4])
            else:  # assume it is a conditional
                conditional = (parts[3], parts[4])
        elif len_parts == 7:
            # the whole sequence, e.g. 1d6 + 3 > 5
            modifier = (parts[3], parts[4])
            conditional = (parts[5], parts[6])
        else:
            # error
            self.caller.msg("You must specify a valid die roll")
            return
        # do the roll
        try:
            result, outcome, diff, rolls = roll_dice(
                ndice, nsides, modifier=modifier, conditional=conditional, return_tuple=True
            )
        except ValueError:
            self.caller.msg(
                "You need to enter valid integer numbers, modifiers and operators."
                " |w%s|n was not understood." % self.args
            )
            return
        # format output
        if len(rolls) > 1:
            rolls = ", ".join(str(roll) for roll in rolls[:-1]) + " and " + str(rolls[-1])
        else:
            rolls = rolls[0]
        if outcome is None:
            outcomestring = ""
        elif outcome:
            outcomestring = " This is a |gsuccess|n (by %s)." % diff
        else:
            outcomestring = " This is a |rfailure|n (by %s)." % diff
        yourollstring = "You roll %s%s."
        roomrollstring = "%s rolls %s%s."
        resultstring = " Roll(s): %s. Total result is |w%s|n."

        if "secret" in self.switches:
            # don't echo to the room at all
            string = yourollstring % (argstring, " (secret, not echoed)")
            string += "\n" + resultstring % (rolls, result)
            string += outcomestring + " (not echoed)"
            self.caller.msg(string)
        elif "hidden" in self.switches:
            # announce the roll to the room, result only to caller
            string = yourollstring % (argstring, " (hidden)")
            self.caller.msg(string)
            string = roomrollstring % (self.caller.key, argstring, " (hidden)")
            self.caller.location.msg_contents(string, exclude=self.caller)
            # handle result
            string = resultstring % (rolls, result)
            string += outcomestring + " (not echoed)"
            self.caller.msg(string)
        else:
            # normal roll
            string = yourollstring % (argstring, "")
            self.caller.msg(string)
            string = roomrollstring % (self.caller.key, argstring, "")
            self.caller.location.msg_contents(string, exclude=self.caller)
            string = resultstring % (rolls, result)
            string += outcomestring
            self.caller.location.msg_contents(string)


class DiceCmdSet(CmdSet):
    """
    a small cmdset for testing purposes.
    Add with @py self.cmdset.add("contrib.dice.DiceCmdSet")
    """

    def at_cmdset_creation(self):
        """Called when set is created"""
        self.add(CmdDice())



# -*- coding: utf-8 -*-

"""
This is part of Evennia's unittest framework, for testing
the stability and integrity of the codebase during updates.

This module tests the lock functionality of Evennia.

"""
from evennia.utils.test_resources import BaseEvenniaTest

try:
    # this is a special optimized Django version, only available in current Django devel
    from django.utils.unittest import TestCase, override_settings
except ImportError:
    from django.test import TestCase, override_settings

from evennia import settings_default
from evennia.locks import lockfuncs
from evennia.utils.create import create_object

# ------------------------------------------------------------
# Lock testing
# ------------------------------------------------------------


class TestLockCheck(BaseEvenniaTest):
    def testrun(self):
        dbref = self.obj2.dbref
        self.obj1.locks.add(
            "owner:dbref(%s);edit:dbref(%s) or perm(Admin);examine:perm(Builder) "
            "and id(%s);delete:perm(Admin);get:all()" % (dbref, dbref, dbref)
        )
        self.obj2.permissions.add("Admin")
        self.assertEqual(True, self.obj1.locks.check(self.obj2, "owner"))
        self.assertEqual(True, self.obj1.locks.check(self.obj2, "edit"))
        self.assertEqual(True, self.obj1.locks.check(self.obj2, "examine"))
        self.assertEqual(True, self.obj1.locks.check(self.obj2, "delete"))
        self.assertEqual(True, self.obj1.locks.check(self.obj2, "get"))
        self.obj1.locks.add("get:false()")
        self.assertEqual(False, self.obj1.locks.check(self.obj2, "get"))
        self.assertEqual(True, self.obj1.locks.check(self.obj2, "not_exist", default=True))


class TestLockfuncs(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        self.account2.permissions.add("Admin")
        self.char2.permissions.add("Builder")

    def test_booleans(self):
        self.assertEqual(True, lockfuncs.true(self.account2, self.obj1))
        self.assertEqual(True, lockfuncs.all(self.account2, self.obj1))
        self.assertEqual(False, lockfuncs.false(self.account2, self.obj1))
        self.assertEqual(False, lockfuncs.none(self.account2, self.obj1))
        self.assertEqual(True, lockfuncs.self(self.obj1, self.obj1))
        self.assertEqual(True, lockfuncs.self(self.account, self.account))
        self.assertEqual(False, lockfuncs.superuser(self.account, None))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_account_perm(self):
        self.assertEqual(False, lockfuncs.perm(self.account2, None, "foo"))
        self.assertEqual(False, lockfuncs.perm(self.account2, None, "Developer"))
        self.assertEqual(False, lockfuncs.perm(self.account2, None, "Developers"))
        self.assertEqual(True, lockfuncs.perm(self.account2, None, "Admin"))
        self.assertEqual(True, lockfuncs.perm(self.account2, None, "Admins"))
        self.assertEqual(True, lockfuncs.perm(self.account2, None, "Player"))
        self.assertEqual(True, lockfuncs.perm(self.account2, None, "Players"))
        self.assertEqual(True, lockfuncs.perm(self.account2, None, "Builder"))
        self.assertEqual(True, lockfuncs.perm(self.account2, None, "Builders"))
        self.assertEqual(True, lockfuncs.perm_above(self.account2, None, "Builder"))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_puppet_perm(self):
        self.assertEqual(False, lockfuncs.perm(self.char2, None, "foo"))
        self.assertEqual(False, lockfuncs.perm(self.char2, None, "Developer"))
        self.assertEqual(False, lockfuncs.perm(self.char2, None, "Develoeprs"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Admin"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Admins"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Player"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Players"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Builder"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Builders"))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_account_perm_above(self):
        self.assertEqual(True, lockfuncs.perm_above(self.char2, None, "Builder"))
        self.assertEqual(True, lockfuncs.perm_above(self.char2, None, "Builders"))
        self.assertEqual(True, lockfuncs.perm_above(self.char2, None, "Player"))
        self.assertEqual(False, lockfuncs.perm_above(self.char2, None, "Admin"))
        self.assertEqual(False, lockfuncs.perm_above(self.char2, None, "Admins"))
        self.assertEqual(False, lockfuncs.perm_above(self.char2, None, "Developers"))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_quell_perm(self):
        self.account2.db._quell = True
        self.assertEqual(False, lockfuncs.false(self.char2, None))
        self.assertEqual(False, lockfuncs.perm(self.char2, None, "Developer"))
        self.assertEqual(False, lockfuncs.perm(self.char2, None, "Developers"))
        self.assertEqual(False, lockfuncs.perm(self.char2, None, "Admin"))
        self.assertEqual(False, lockfuncs.perm(self.char2, None, "Admins"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Player"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Players"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Builder"))
        self.assertEqual(True, lockfuncs.perm(self.char2, None, "Builders"))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_quell_above_perm(self):
        self.assertEqual(True, lockfuncs.perm_above(self.char2, None, "Player"))
        self.assertEqual(True, lockfuncs.perm_above(self.char2, None, "Builder"))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_object_perm(self):
        self.obj2.permissions.add("Admin")
        self.assertEqual(False, lockfuncs.perm(self.obj2, None, "Developer"))
        self.assertEqual(False, lockfuncs.perm(self.obj2, None, "Developers"))
        self.assertEqual(True, lockfuncs.perm(self.obj2, None, "Admin"))
        self.assertEqual(True, lockfuncs.perm(self.obj2, None, "Admins"))
        self.assertEqual(True, lockfuncs.perm(self.obj2, None, "Player"))
        self.assertEqual(True, lockfuncs.perm(self.obj2, None, "Players"))
        self.assertEqual(True, lockfuncs.perm(self.obj2, None, "Builder"))
        self.assertEqual(True, lockfuncs.perm(self.obj2, None, "Builders"))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_object_above_perm(self):
        self.obj2.permissions.add("Admin")
        self.assertEqual(False, lockfuncs.perm_above(self.obj2, None, "Admins"))
        self.assertEqual(True, lockfuncs.perm_above(self.obj2, None, "Builder"))
        self.assertEqual(True, lockfuncs.perm_above(self.obj2, None, "Builders"))

    @override_settings(PERMISSION_HIERARCHY=settings_default.PERMISSION_HIERARCHY)
    def test_pperm(self):
        self.obj2.permissions.add("Developer")
        self.char2.permissions.add("Developer")
        self.assertEqual(False, lockfuncs.pperm(self.obj2, None, "Players"))
        self.assertEqual(True, lockfuncs.pperm(self.char2, None, "Players"))
        self.assertEqual(True, lockfuncs.pperm(self.account, None, "Admins"))
        self.assertEqual(True, lockfuncs.pperm_above(self.account, None, "Builders"))
        self.assertEqual(False, lockfuncs.pperm_above(self.account2, None, "Admins"))
        self.assertEqual(True, lockfuncs.pperm_above(self.char2, None, "Players"))

    def test_dbref(self):
        dbref = self.obj2.dbref
        self.assertEqual(True, lockfuncs.dbref(self.obj2, None, "%s" % dbref))
        self.assertEqual(False, lockfuncs.id(self.obj2, None, "%s" % (dbref + "1")))
        dbref = self.account2.dbref
        self.assertEqual(True, lockfuncs.pdbref(self.account2, None, "%s" % dbref))
        self.assertEqual(False, lockfuncs.pid(self.account2, None, "%s" % (dbref + "1")))

    def test_attr(self):
        self.obj2.db.testattr = 45
        self.assertEqual(True, lockfuncs.attr(self.obj2, None, "testattr", "45"))
        self.assertEqual(False, lockfuncs.attr_gt(self.obj2, None, "testattr", "45"))
        self.assertEqual(True, lockfuncs.attr_ge(self.obj2, None, "testattr", "45"))
        self.assertEqual(False, lockfuncs.attr_lt(self.obj2, None, "testattr", "45"))
        self.assertEqual(True, lockfuncs.attr_le(self.obj2, None, "testattr", "45"))

        self.assertEqual(True, lockfuncs.objattr(None, self.obj2, "testattr", "45"))
        self.assertEqual(True, lockfuncs.objattr(None, self.obj2, "testattr", "45"))
        self.assertEqual(False, lockfuncs.objattr(None, self.obj2, "testattr", "45", compare="lt"))

    def test_locattr(self):
        self.obj2.location.db.locattr = "test"
        self.assertEqual(True, lockfuncs.locattr(self.obj2, None, "locattr", "test"))
        self.assertEqual(False, lockfuncs.locattr(self.obj2, None, "fail", "testfail"))
        self.assertEqual(True, lockfuncs.objlocattr(None, self.obj2, "locattr", "test"))

    def test_tag(self):
        self.obj2.tags.add("test1")
        self.obj2.tags.add("test2", "category1")
        self.assertEqual(True, lockfuncs.tag(self.obj2, None, "test1"))
        self.assertEqual(True, lockfuncs.tag(self.obj2, None, "test2", "category1"))
        self.assertEqual(False, lockfuncs.tag(self.obj2, None, "test1", "category1"))
        self.assertEqual(False, lockfuncs.tag(self.obj2, None, "test1", "category2"))
        self.assertEqual(True, lockfuncs.objtag(None, self.obj2, "test2", "category1"))
        self.assertEqual(False, lockfuncs.objtag(None, self.obj2, "test2"))

    def test_traverse_taglock(self):
        self.obj2.tags.add("test1", "category1")
        self.exit.locks.add("traverse:tag(test1,category1)")
        self.assertEqual(self.exit.access(self.obj2, "traverse"), True)

    def test_traverse_taglock_fail(self):
        self.obj2.tags.add("test1")  # missing the category
        self.exit.locks.add("traverse:tag(test1,category1)")
        self.assertEqual(self.exit.access(self.obj2, "traverse"), False)

    def test_inside_holds(self):
        self.assertEqual(True, lockfuncs.inside(self.char1, self.room1))
        self.assertEqual(False, lockfuncs.inside(self.char1, self.room2))
        self.assertEqual(True, lockfuncs.holds(self.room1, self.char1))
        self.assertEqual(False, lockfuncs.holds(self.room2, self.char1))
        # test recursively
        self.assertEqual(True, lockfuncs.inside_rec(self.char1, self.room1))
        self.assertEqual(False, lockfuncs.inside_rec(self.char1, self.room2))
        inventory_item = create_object(key="InsideTester", location=self.char1)
        self.assertEqual(True, lockfuncs.inside_rec(inventory_item, self.room1))
        self.assertEqual(False, lockfuncs.inside_rec(inventory_item, self.room2))
        inventory_item.delete()

    def test_has_account(self):
        self.assertEqual(True, lockfuncs.has_account(self.char1, None))
        self.assertEqual(False, lockfuncs.has_account(self.obj1, None))

    @override_settings(IRC_ENABLED=True, TESTVAL=[1, 2, 3])
    def test_serversetting(self):
        self.assertEqual(True, lockfuncs.serversetting(None, None, "IRC_ENABLED", "True"))
        self.assertEqual(True, lockfuncs.serversetting(None, None, "TESTVAL", "[1, 2, 3]"))
        self.assertEqual(False, lockfuncs.serversetting(None, None, "TESTVAL", "[1, 2, 4]"))
        self.assertEqual(False, lockfuncs.serversetting(None, None, "TESTVAL", "123"))

    def test_is_ooc__char(self):
        self.assertEqual(False, lockfuncs.is_ooc(self.char1, self.char1))

    def test_is_ooc__session(self):
        self.assertEqual(False, lockfuncs.is_ooc(self.session, self.char1))

    def test_is_ooc__account(self):
        self.assertEqual(False, lockfuncs.is_ooc(self.account, self.char1))
        self.account.unpuppet_all()
        self.assertEqual(True, lockfuncs.is_ooc(self.account, self.char1))


class TestPermissionCheck(BaseEvenniaTest):
    """
    Test the PermissionHandler.check method

    """

    def test_check__success(self):
        """Test combinations that should pass the check"""
        self.assertEqual(
            [perm for perm in self.char1.account.permissions.all()], ["developer", "player"]
        )
        self.assertTrue(self.char1.permissions.check("Builder"))
        self.assertTrue(self.char1.permissions.check("Builder", "Player"))
        self.assertTrue(self.char1.permissions.check("Builder", "dummy"))
        self.assertTrue(self.char1.permissions.check("Developer", "dummy", "foobar"))
        self.assertTrue(self.char1.permissions.check("Builder", "Player", require_all=True))

    def test_check__fail(self):
        """Test combinations that should fail the check"""
        self.assertFalse(self.char1.permissions.check("dummy"))
        self.assertFalse(self.char1.permissions.check("Builder", "dummy", require_all=True))
        self.assertFalse(self.char1.permissions.check("Developer", "foobar", require_all=True))

        self.char1.account.permissions.remove("developer")
        self.char1.account.permissions.add("Builder")

        self.assertEqual(
            [perm for perm in self.char1.account.permissions.all()], ["builder", "player"]
        )
        self.assertFalse(self.char1.permissions.check("Developer"))
        self.assertFalse(self.char1.permissions.check("Developer", "Player", require_all=True))
        self.assertFalse(self.char1.permissions.check("Player", "dummy", require_all=True))



from evennia.objects.objects import DefaultCharacter
from evennia.utils import create
from evennia.utils.test_resources import BaseEvenniaTest, EvenniaTest

from . import signals
from .component import Component
from .dbfield import DBField, TagField
from .holder import ComponentHolderMixin, ComponentProperty
from .signals import as_listener


class ComponentTestA(Component):
    name = "test_a"
    my_int = DBField(default=1)
    my_list = DBField(default=[], autocreate=True)


class ShadowedComponentTestA(ComponentTestA):
    name = "shadowed_test_a"
    slot = "ic_a"


class InheritedComponentTestA(ComponentTestA):
    name = "inherited_test_a"
    slot = "ic_a"

    my_other_int = DBField(default=2)


class ReplacementComponentTestA(InheritedComponentTestA):
    name = "replacement_inherited_test_a"
    slot = "ic_a"

    replacement_field = DBField(default=6)


class ComponentTestB(Component):
    name = "test_b"
    my_int = DBField(default=1)
    my_list = DBField(default=[], autocreate=True)
    default_tag = TagField(default="initial_value")
    single_tag = TagField(enforce_single=True)
    multiple_tags = TagField()
    default_single_tag = TagField(default="initial_value", enforce_single=True)


class RuntimeComponentTestC(Component):
    name = "test_c"
    my_int = DBField(default=6)
    my_dict = DBField(default={}, autocreate=True)
    added_tag = TagField(default="added_value")


class ComponentTestD(Component):
    name = "test_d"

    mixed_in = DBField(default=8)


class ShadowedCharacterMixin:
    ic_a = ComponentProperty("shadowed_test_a")


class CharacterMixinWithComponents:
    ic_a = ComponentProperty("inherited_test_a", my_other_int=33)
    test_d = ComponentProperty("test_d")


class CharacterWithComponents(
    ComponentHolderMixin, ShadowedCharacterMixin, CharacterMixinWithComponents, DefaultCharacter
):
    test_a = ComponentProperty("test_a")
    test_b = ComponentProperty("test_b", my_int=3, my_list=[1, 2, 3])
    ic_a = ComponentProperty("inherited_test_a", my_other_int=4)


class InheritedTCWithComponents(CharacterWithComponents):
    test_c = ComponentProperty("test_c")


class TestComponents(EvenniaTest):
    character_typeclass = CharacterWithComponents

    def test_character_has_class_components(self):
        self.assertTrue(self.char1.test_a)
        self.assertTrue(self.char1.test_b)

    def test_inherited_typeclass_does_not_include_child_class_components(self):
        char_with_c = create.create_object(
            InheritedTCWithComponents, key="char_with_c", location=self.room1, home=self.room1
        )
        self.assertTrue(self.char1.test_a)
        self.assertFalse(self.char1.cmp.get("test_c"))
        self.assertTrue(char_with_c.test_c)

    def test_character_instances_components_properly(self):
        self.assertIsInstance(self.char1.test_a, ComponentTestA)
        self.assertIsInstance(self.char1.test_b, ComponentTestB)

    def test_character_assigns_default_value(self):
        self.assertEquals(self.char1.test_a.my_int, 1)
        self.assertEquals(self.char1.test_a.my_list, [])

    def test_character_assigns_default_provided_values(self):
        self.assertEquals(self.char1.test_b.my_int, 3)
        self.assertEquals(self.char1.test_b.my_list, [1, 2, 3])

    def test_character_has_autocreated_values(self):
        att_name = "test_b::my_list"
        self.assertEquals(self.char1.attributes.get(att_name), [1, 2, 3])

    def test_component_inheritance_properly_overrides_slots(self):
        self.assertEquals(self.char1.ic_a.name, "inherited_test_a")
        component_names = set(c[0] for c in self.char1._get_class_components())
        self.assertNotIn("shadowed_test_a", component_names)

    def test_component_inheritance_assigns_proper_values(self):
        self.assertEquals(self.char1.ic_a.my_int, 1)
        self.assertEquals(self.char1.ic_a.my_other_int, 4)

    def test_host_mixins_assigns_components(self):
        self.assertEquals(self.char1.test_d.mixed_in, 8)

    def test_character_can_register_runtime_component(self):
        rct = RuntimeComponentTestC.create(self.char1)
        self.char1.components.add(rct)
        test_c = self.char1.components.get("test_c")

        self.assertTrue(test_c)
        self.assertEquals(test_c.my_int, 6)
        self.assertEquals(test_c.my_dict, {})

    def test_handler_can_add_default_component(self):
        self.char1.components.add_default("test_c")
        test_c = self.char1.components.get("test_c")

        self.assertTrue(test_c)
        self.assertEquals(test_c.my_int, 6)

    def test_handler_has_returns_true_for_any_components(self):
        rct = RuntimeComponentTestC.create(self.char1)
        handler = self.char1.components
        handler.add(rct)

        self.assertTrue(handler.has("test_a"))
        self.assertTrue(handler.has("test_b"))
        self.assertTrue(handler.has("test_c"))

    def test_can_remove_component(self):
        rct = RuntimeComponentTestC.create(self.char1)
        handler = self.char1.components
        handler.add(rct)
        handler.remove(rct)

        self.assertTrue(handler.has("test_a"))
        self.assertTrue(handler.has("test_b"))
        self.assertFalse(handler.has("test_c"))

    def test_can_remove_component_by_name(self):
        rct = RuntimeComponentTestC.create(self.char1)
        handler = self.char1.components
        handler.add(rct)
        handler.remove_by_name("test_c")

        self.assertTrue(handler.has("test_a"))
        self.assertTrue(handler.has("test_b"))
        self.assertFalse(handler.has("test_c"))

    def test_cannot_replace_component(self):
        with self.assertRaises(Exception):
            self.char1.test_a = None

    def test_can_get_component(self):
        rct = RuntimeComponentTestC.create(self.char1)
        handler = self.char1.components
        handler.add(rct)

        self.assertIs(handler.get("test_c"), rct)

    def test_can_access_component_regular_get(self):
        self.assertIs(self.char1.cmp.test_a, self.char1.components.get("test_a"))

    def test_returns_none_with_regular_get_when_no_attribute(self):
        self.assertIs(self.char1.cmp.does_not_exist, None)

    def test_host_has_class_component_tags(self):
        self.assertTrue(self.char1.tags.has(key="test_a", category="components"))
        self.assertTrue(self.char1.tags.has(key="test_b", category="components"))
        self.assertTrue(self.char1.tags.has(key="initial_value", category="test_b::default_tag"))
        self.assertTrue(self.char1.test_b.default_tag == "initial_value")
        self.assertFalse(self.char1.tags.has(key="test_c", category="components"))
        self.assertFalse(self.char1.tags.has(category="test_b::single_tag"))
        self.assertFalse(self.char1.tags.has(category="test_b::multiple_tags"))

    def test_host_has_added_component_tags(self):
        rct = RuntimeComponentTestC.create(self.char1)
        self.char1.components.add(rct)
        test_c = self.char1.components.get("test_c")

        self.assertTrue(self.char1.tags.has(key="test_c", category="components"))
        self.assertTrue(self.char1.tags.has(key="added_value", category="test_c::added_tag"))
        self.assertEquals(test_c.added_tag, "added_value")

    def test_host_has_added_default_component_tags(self):
        self.char1.components.add_default("test_c")
        test_c = self.char1.components.get("test_c")

        self.assertTrue(self.char1.tags.has(key="test_c", category="components"))
        self.assertTrue(self.char1.tags.has(key="added_value", category="test_c::added_tag"))
        self.assertEquals(test_c.added_tag, "added_value")

    def test_host_remove_component_tags(self):
        rct = RuntimeComponentTestC.create(self.char1)
        handler = self.char1.components
        handler.add(rct)
        self.assertTrue(self.char1.tags.has(key="test_c", category="components"))
        handler.remove(rct)

        self.assertFalse(self.char1.tags.has(key="test_c", category="components"))
        self.assertFalse(self.char1.tags.has(key="added_value", category="test_c::added_tag"))

    def test_host_remove_by_name_component_tags(self):
        rct = RuntimeComponentTestC.create(self.char1)
        handler = self.char1.components
        handler.add(rct)
        self.assertTrue(self.char1.tags.has(key="test_c", category="components"))
        handler.remove_by_name("test_c")

        self.assertFalse(self.char1.tags.has(key="test_c", category="components"))
        self.assertFalse(self.char1.tags.has(key="added_value", category="test_c::added_tag"))

    def test_component_tags_only_hold_one_value_when_enforce_single(self):
        test_b = self.char1.components.get("test_b")
        test_b.single_tag = "first_value"
        test_b.single_tag = "second value"

        self.assertTrue(self.char1.tags.has(key="second value", category="test_b::single_tag"))
        self.assertEquals(test_b.single_tag, "second value")
        self.assertFalse(self.char1.tags.has(key="first_value", category="test_b::single_tag"))

    def test_component_tags_default_value_is_overridden_when_enforce_single(self):
        test_b = self.char1.components.get("test_b")
        test_b.default_single_tag = "second value"

        self.assertTrue(
            self.char1.tags.has(key="second value", category="test_b::default_single_tag")
        )
        self.assertTrue(test_b.default_single_tag == "second value")
        self.assertFalse(
            self.char1.tags.has(key="first_value", category="test_b::default_single_tag")
        )

    def test_component_tags_support_multiple_values_by_default(self):
        test_b = self.char1.components.get("test_b")
        test_b.multiple_tags = "first value"
        test_b.multiple_tags = "second value"
        test_b.multiple_tags = "third value"

        self.assertTrue(
            all(
                val in test_b.multiple_tags
                for val in ("first value", "second value", "third value")
            )
        )
        self.assertTrue(self.char1.tags.has(key="first value", category="test_b::multiple_tags"))
        self.assertTrue(self.char1.tags.has(key="second value", category="test_b::multiple_tags"))
        self.assertTrue(self.char1.tags.has(key="third value", category="test_b::multiple_tags"))

    def test_mutables_are_not_shared_when_autocreate(self):
        self.char1.test_a.my_list.append(1)
        self.assertNotEquals(self.char1.test_a.my_list, self.char2.test_a.my_list)

    def test_replacing_class_component_slot_with_runtime_component(self):
        self.char1.components.add_default("replacement_inherited_test_a")
        self.assertEquals(self.char1.ic_a.replacement_field, 6)


class CharWithSignal(ComponentHolderMixin, DefaultCharacter):
    @signals.as_listener
    def my_signal(self):
        setattr(self, "my_signal_is_called", True)

    @signals.as_listener
    def my_other_signal(self):
        setattr(self, "my_other_signal_is_called", True)

    @signals.as_responder
    def my_response(self):
        return 1

    @signals.as_responder
    def my_other_response(self):
        return 2


class ComponentWithSignal(Component):
    name = "test_signal_a"

    @signals.as_listener
    def my_signal(self):
        setattr(self, "my_signal_is_called", True)

    @signals.as_listener
    def my_other_signal(self):
        setattr(self, "my_other_signal_is_called", True)

    @signals.as_responder
    def my_response(self):
        return 1

    @signals.as_responder
    def my_other_response(self):
        return 2

    @signals.as_responder
    def my_component_response(self):
        return 3


class TestComponentSignals(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        self.char1 = create.create_object(
            CharWithSignal,
            key="Char",
        )

    def test_host_can_register_as_listener(self):
        self.char1.signals.trigger("my_signal")

        self.assertTrue(self.char1.my_signal_is_called)
        self.assertFalse(getattr(self.char1, "my_other_signal_is_called", None))

    def test_host_can_register_as_responder(self):
        responses = self.char1.signals.query("my_response")

        self.assertIn(1, responses)
        self.assertNotIn(2, responses)

    def test_component_can_register_as_listener(self):
        char = self.char1
        char.components.add(ComponentWithSignal.create(char))
        char.signals.trigger("my_signal")

        component = char.cmp.test_signal_a
        self.assertTrue(component.my_signal_is_called)
        self.assertFalse(getattr(component, "my_other_signal_is_called", None))

    def test_component_can_register_as_responder(self):
        char = self.char1
        char.components.add(ComponentWithSignal.create(char))
        responses = char.signals.query("my_response")

        self.assertIn(1, responses)
        self.assertNotIn(2, responses)

    def test_signals_can_add_listener(self):
        result = []

        def my_fake_listener():
            result.append(True)

        self.char1.signals.add_listener("my_fake_signal", my_fake_listener)
        self.char1.signals.trigger("my_fake_signal")

        self.assertTrue(result)

    def test_signals_can_add_responder(self):
        def my_fake_responder():
            return 1

        self.char1.signals.add_responder("my_fake_response", my_fake_responder)
        responses = self.char1.signals.query("my_fake_response")

        self.assertIn(1, responses)

    def test_signals_can_remove_listener(self):
        result = []

        def my_fake_listener():
            result.append(True)

        self.char1.signals.add_listener("my_fake_signal", my_fake_listener)
        self.char1.signals.remove_listener("my_fake_signal", my_fake_listener)
        self.char1.signals.trigger("my_fake_signal")

        self.assertFalse(result)

    def test_signals_can_remove_responder(self):
        def my_fake_responder():
            return 1

        self.char1.signals.add_responder("my_fake_response", my_fake_responder)
        self.char1.signals.remove_responder("my_fake_response", my_fake_responder)
        responses = self.char1.signals.query("my_fake_response")

        self.assertFalse(responses)

    def test_signals_can_trigger_with_args(self):
        result = []

        def my_fake_listener(arg1, kwarg1):
            result.append((arg1, kwarg1))

        self.char1.signals.add_listener("my_fake_signal", my_fake_listener)
        self.char1.signals.trigger("my_fake_signal", 1, kwarg1=2)

        self.assertIn((1, 2), result)

    def test_signals_can_query_with_args(self):
        def my_fake_responder(arg1, kwarg1):
            return (arg1, kwarg1)

        self.char1.signals.add_responder("my_fake_response", my_fake_responder)
        responses = self.char1.signals.query("my_fake_response", 1, kwarg1=2)

        self.assertIn((1, 2), responses)

    def test_signals_trigger_does_not_fail_without_listener(self):
        self.char1.signals.trigger("some_unknown_signal")

    def test_signals_query_does_not_fail_wihout_responders(self):
        self.char1.signals.query("no_responders_allowed")

    def test_signals_query_with_aggregate(self):
        def my_fake_responder(arg1, kwarg1):
            return (arg1, kwarg1)

        self.char1.signals.add_responder("my_fake_response", my_fake_responder)
        responses = self.char1.signals.query("my_fake_response", 1, kwarg1=2)

        self.assertIn((1, 2), responses)

    def test_signals_can_add_object_listeners_and_responders(self):
        result = []

        class FakeObj:
            @as_listener
            def my_signal(self):
                result.append(True)

        self.char1.signals.add_object_listeners_and_responders(FakeObj())
        self.char1.signals.trigger("my_signal")

        self.assertTrue(result)

    def test_signals_can_remove_object_listeners_and_responders(self):
        result = []

        class FakeObj:
            @as_listener
            def my_signal(self):
                result.append(True)

        obj = FakeObj()
        self.char1.signals.add_object_listeners_and_responders(obj)
        self.char1.signals.remove_object_listeners_and_responders(obj)
        self.char1.signals.trigger("my_signal")

        self.assertFalse(result)

    def test_component_handler_signals_connected_when_adding_default_component(self):
        char = self.char1
        char.components.add_default("test_signal_a")
        responses = char.signals.query("my_component_response")

        self.assertIn(3, responses)

    def test_component_handler_signals_disconnected_when_removing_component(self):
        char = self.char1
        comp = ComponentWithSignal.create(char)
        char.components.add(comp)
        char.components.remove(comp)
        responses = char.signals.query("my_component_response")

        self.assertFalse(responses)

    def test_component_handler_signals_disconnected_when_removing_component_by_name(self):
        char = self.char1
        char.components.add_default("test_signal_a")
        char.components.remove_by_name("test_signal_a")
        responses = char.signals.query("my_component_response")

        self.assertFalse(responses)



"""
Test the contrib barter system
"""

from mock import Mock

from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.utils.create import create_object

from . import barter


class TestBarter(BaseEvenniaCommandTest):
    def setUp(self):
        super().setUp()
        self.tradeitem1 = create_object(key="TradeItem1", location=self.char1)
        self.tradeitem2 = create_object(key="TradeItem2", location=self.char1)
        self.tradeitem3 = create_object(key="TradeItem3", location=self.char2)

    def test_tradehandler_base(self):
        self.char1.msg = Mock()
        self.char2.msg = Mock()
        # test all methods of the tradehandler
        handler = barter.TradeHandler(self.char1, self.char2)
        self.assertEqual(handler.part_a, self.char1)
        self.assertEqual(handler.part_b, self.char2)
        handler.msg_other(self.char1, "Want to trade?")
        handler.msg_other(self.char2, "Yes!")
        handler.msg_other(None, "Talking to myself...")
        self.assertEqual(self.char2.msg.mock_calls[0][1][0], "Want to trade?")
        self.assertEqual(self.char1.msg.mock_calls[0][1][0], "Yes!")
        self.assertEqual(self.char1.msg.mock_calls[1][1][0], "Talking to myself...")
        self.assertEqual(handler.get_other(self.char1), self.char2)
        handler.finish(force=True)

    def test_tradehandler_joins(self):
        handler = barter.TradeHandler(self.char1, self.char2)
        self.assertTrue(handler.join(self.char2))
        self.assertTrue(handler.unjoin(self.char2))
        self.assertFalse(handler.join(self.char1))
        self.assertFalse(handler.unjoin(self.char1))
        handler.finish(force=True)

    def test_tradehandler_offers(self):
        handler = barter.TradeHandler(self.char1, self.char2)
        handler.join(self.char2)
        handler.offer(self.char1, self.tradeitem1, self.tradeitem2)
        self.assertEqual(handler.part_a_offers, [self.tradeitem1, self.tradeitem2])
        self.assertFalse(handler.part_a_accepted)
        self.assertFalse(handler.part_b_accepted)
        handler.offer(self.char2, self.tradeitem3)
        self.assertEqual(handler.list(), ([self.tradeitem1, self.tradeitem2], [self.tradeitem3]))
        self.assertEqual(handler.search("TradeItem2"), self.tradeitem2)
        self.assertEqual(handler.search("TradeItem3"), self.tradeitem3)
        self.assertEqual(handler.search("nonexisting"), None)
        self.assertFalse(handler.finish())  # should fail since offer not yet accepted
        handler.accept(self.char1)
        handler.decline(self.char1)
        handler.accept(self.char2)
        handler.accept(self.char1)  # should trigger handler.finish() automatically
        self.assertEqual(self.tradeitem1.location, self.char2)
        self.assertEqual(self.tradeitem2.location, self.char2)
        self.assertEqual(self.tradeitem3.location, self.char1)

    def test_cmdtrade(self):
        self.call(
            barter.CmdTrade(),
            "Char2 : Hey wanna trade?",
            'You say, "Hey wanna trade?"',
            caller=self.char1,
        )
        self.call(barter.CmdTrade(), "Char decline : Nope!", 'You say, "Nope!"', caller=self.char2)
        self.call(
            barter.CmdTrade(),
            "Char2 : Hey wanna trade?",
            'You say, "Hey wanna trade?"',
            caller=self.char1,
        )
        self.call(barter.CmdTrade(), "Char accept : Sure!", 'You say, "Sure!"', caller=self.char2)
        self.call(
            barter.CmdOffer(),
            "TradeItem3",
            "Your trade action: You offer TradeItem3",
            caller=self.char2,
        )
        self.call(
            barter.CmdOffer(),
            "TradeItem1 : Here's my offer.",
            'You say, "Here\'s my offer."\n  [You offer TradeItem1]',
        )
        self.call(
            barter.CmdAccept(),
            "",
            "Your trade action: You accept the offer. Char2 must now also accept",
        )
        self.call(
            barter.CmdDecline(),
            "",
            "Your trade action: You change your mind, declining the current offer.",
        )
        self.call(
            barter.CmdAccept(),
            ": Sounds good.",
            'You say, "Sounds good."\n' "  [You accept the offer. Char must now also accept.",
            caller=self.char2,
        )
        self.call(
            barter.CmdDecline(),
            ":No way!",
            'You say, "No way!"\n  [You change your mind, declining the current offer.]',
            caller=self.char2,
        )
        self.call(
            barter.CmdOffer(),
            "TradeItem1, TradeItem2 : My final offer!",
            'You say, "My final offer!"\n  [You offer TradeItem1 and TradeItem2]',
        )
        self.call(
            barter.CmdAccept(),
            "",
            "Your trade action: You accept the offer. Char2 must now also accept.",
            caller=self.char1,
        )
        self.call(barter.CmdStatus(), "", "Offered by Char:", caller=self.char2)
        self.tradeitem1.db.desc = "A great offer."
        self.call(barter.CmdEvaluate(), "TradeItem1", "A great offer.")
        self.call(
            barter.CmdAccept(),
            ":Ok then.",
            'You say, "Ok then."\n  [You accept the deal.',
            caller=self.char2,
        )
        self.assertEqual(self.tradeitem1.location, self.char2)
        self.assertEqual(self.tradeitem2.location, self.char2)
        self.assertEqual(self.tradeitem3.location, self.char1)

    def test_cmdtradehelp(self):
        self.call(
            barter.CmdTrade(),
            "Char2 : Hey wanna trade?",
            'You say, "Hey wanna trade?"',
            caller=self.char1,
        )
        self.call(barter.CmdTradeHelp(), "", "Trading commands\n", caller=self.char1)
        self.call(
            barter.CmdFinish(),
            ": Ending.",
            'You say, "Ending."\n  [You aborted trade. No deal was made.]',
        )



"""
Test of the Unixcommand.

"""

from evennia.commands.default.tests import BaseEvenniaCommandTest

from .unixcommand import UnixCommand


class CmdDummy(UnixCommand):

    """A dummy UnixCommand."""

    key = "dummy"

    def init_parser(self):
        """Fill out options."""
        self.parser.add_argument("nb1", type=int, help="the first number")
        self.parser.add_argument("nb2", type=int, help="the second number")
        self.parser.add_argument("-v", "--verbose", action="store_true")

    def func(self):
        nb1 = self.opts.nb1
        nb2 = self.opts.nb2
        result = nb1 * nb2
        verbose = self.opts.verbose
        if verbose:
            self.msg("{} times {} is {}".format(nb1, nb2, result))
        else:
            self.msg("{} * {} = {}".format(nb1, nb2, result))


class TestUnixCommand(BaseEvenniaCommandTest):
    def test_success(self):
        """See the command parsing succeed."""
        self.call(CmdDummy(), "5 10", "5 * 10 = 50")
        self.call(CmdDummy(), "5 10 -v", "5 times 10 is 50")

    def test_failure(self):
        """If not provided with the right info, should fail."""
        ret = self.call(CmdDummy(), "5")
        lines = ret.splitlines()
        self.assertTrue(any(lin.startswith("usage:") for lin in lines))
        self.assertTrue(any(lin.startswith("dummy: error:") for lin in lines))

        # If we specify an incorrect number as parameter
        ret = self.call(CmdDummy(), "five ten")
        lines = ret.splitlines()
        self.assertTrue(any(lin.startswith("usage:") for lin in lines))
        self.assertTrue(any(lin.startswith("dummy: error:") for lin in lines))



"""
Containers

Containers are storage classes usually initialized from a setting. They
represent Singletons and acts as a convenient place to find resources (
available as properties on the singleton)

evennia.GLOBAL_SCRIPTS
evennia.OPTION_CLASSES

"""


from pickle import dumps

from django.conf import settings
from django.db.utils import OperationalError, ProgrammingError

from evennia.utils import logger
from evennia.utils.utils import callables_from_module, class_from_module

SCRIPTDB = None


class Container:
    """
    Base container class. A container is simply a storage object whose
    properties can be acquired as a property on it. This is generally
    considered a read-only affair.

    The container is initialized by a list of modules containing callables.

    """

    storage_modules = []

    def __init__(self):
        """
        Read data from module.

        """
        self.loaded_data = None

    def load_data(self):
        """
        Delayed import to avoid eventual circular imports from inside
        the storage modules.

        """
        if self.loaded_data is None:
            self.loaded_data = {}
            for module in self.storage_modules:
                self.loaded_data.update(callables_from_module(module))

    def __getattr__(self, key):
        return self.get(key)

    def get(self, key, default=None):
        """
        Retrive data by key (in case of not knowing it beforehand).

        Args:
            key (str): The name of the script.
            default (any, optional): Value to return if key is not found.

        Returns:
            any (any): The data loaded on this container.

        """
        self.load_data()
        return self.loaded_data.get(key, default)

    def all(self):
        """
        Get all stored data

        Returns:
            scripts (list): All global script objects stored on the container.

        """
        self.load_data()
        return list(self.loaded_data.values())


class OptionContainer(Container):
    """
    Loads and stores the final list of OPTION CLASSES.

    Can access these as properties or dictionary-contents.
    """

    storage_modules = settings.OPTION_CLASS_MODULES


class GlobalScriptContainer(Container):
    """
    Simple Handler object loaded by the Evennia API to contain and manage a
    game's Global Scripts. This will list global Scripts created on their own
    but will also auto-(re)create scripts defined in `settings.GLOBAL_SCRIPTS`.

    Example:
        import evennia
        evennia.GLOBAL_SCRIPTS.scriptname

    Note:
        This does not use much of the BaseContainer since it's not loading
        callables from settings but a custom dict of tuples.

    """

    def __init__(self):
        """
        Note: We must delay loading of typeclasses since this module may get
        initialized before Scripts are actually initialized.

        """
        self.typeclass_storage = dict()
        self.loaded_data = {
            key: {} if data is None else data for key, data in settings.GLOBAL_SCRIPTS.items()
        }
        self.loaded = False

    def _load_script(self, key):
        typeclass = self.typeclass_storage[key]
        script = typeclass.objects.filter(
            db_key=key, db_account__isnull=True, db_obj__isnull=True
        ).first()

        kwargs = {**self.loaded_data[key]}
        kwargs["key"] = key
        kwargs["persistent"] = kwargs.get("persistent", True)

        compare_hash = str(dumps(kwargs, protocol=4))

        if script:
            script_hash = script.attributes.get("global_script_settings", category="settings_hash")
            if script_hash is None:
                # legacy - store the hash anew and assume no change
                script.attributes.add(
                    "global_script_settings", compare_hash, category="settings_hash"
                )
            elif script_hash != compare_hash:
                # wipe the old version and create anew
                logger.log_info(f"GLOBAL_SCRIPTS: Settings changed for {key} ({typeclass}).")
                script.stop()
                script.delete()
                script = None

        if not script:
            logger.log_info(f"GLOBAL_SCRIPTS: (Re)creating {key} ({typeclass}).")

            script, errors = typeclass.create(**kwargs)
            if errors:
                logger.log_err("\n".join(errors))
                return None

            # store a hash representation of the setup
            script.attributes.add("global_script_settings", compare_hash, category="settings_hash")

        return script

    def start(self):
        """
        Called last in evennia.__init__ to initialize the container late
        (after script typeclasses have finished loading).

        We include all global scripts in the handler and
        make sure to auto-load time-based scripts.

        """
        # populate self.typeclass_storage
        if not self.loaded:
            self.load_data()

        # make sure settings-defined scripts are loaded
        scripts_to_run = []
        for key in self.loaded_data:
            script = self._load_script(key)
            if script:
                scripts_to_run.append(script)
        # start all global scripts
        try:
            for script in scripts_to_run:
                script.start()
        except (OperationalError, ProgrammingError):
            # this can happen if db is not loaded yet (such as when building docs)
            pass

    def load_data(self):
        """
        This delayed import avoids trying to load Scripts before they are
        initialized.

        """
        if self.loaded:
            return
        if not self.typeclass_storage:
            for key, data in list(self.loaded_data.items()):
                typeclass = data.get("typeclass", settings.BASE_SCRIPT_TYPECLASS)
                self.typeclass_storage[key] = class_from_module(
                    typeclass, fallback=settings.BASE_SCRIPT_TYPECLASS
                )
        self.loaded = True

    def get(self, key, default=None):
        """
        Retrive data by key (in case of not knowing it beforehand). Any
        scripts that are in settings.GLOBAL_SCRIPTS that are not found
        will be recreated on-demand.

        Args:
            key (str): The name of the script.
            default (any, optional): Value to return if key is not found
                at all on this container (i.e it cannot be loaded at all).

        Returns:
            any (any): The data loaded on this container.
        """
        if not self.loaded:
            self.load_data()
        out_value = default
        if key in self.loaded_data:
            if key not in self.typeclass_storage:
                # this means we are trying to load in a loop
                raise RuntimeError(
                    f"Trying to access `GLOBAL_SCRIPTS.{key}` before scripts have finished "
                    "initializing. This can happen if accessing GLOBAL_SCRIPTS from the same "
                    "module the script is defined in."
                )
            # recreate if we have the info
            script_found = self._load_script(key)
            if script_found:
                out_value = script_found

        return out_value

    def all(self):
        """
        Get all global scripts. Note that this will not auto-start
        scripts defined in settings.

        Returns:
            scripts (list): All global script objects stored on the container.

        """
        if not self.loaded:
            self.load_data()
        return self.scripts.values()


# Create all singletons

GLOBAL_SCRIPTS = GlobalScriptContainer()
OPTION_CLASSES = OptionContainer()



from evennia import create_object
from evennia.utils.test_resources import BaseEvenniaCommandTest, BaseEvenniaTest  # noqa

from .containers import CmdContainerGet, CmdContainerLook, CmdPut, ContribContainer


class TestContainer(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        # create a container to test with
        self.container = create_object(key="Box", typeclass=ContribContainer, location=self.room1)

    def test_capacity(self):
        # limit capacity to 1
        self.container.capacity = 1
        self.assertTrue(self.container.at_pre_put_in(self.char1, self.obj1))
        # put Obj2 in container to hit max capacity
        self.obj2.location = self.container
        self.assertFalse(self.container.at_pre_put_in(self.char1, self.obj1))


class TestContainerCmds(BaseEvenniaCommandTest):
    def setUp(self):
        super().setUp()
        # create a container to test with
        self.container = create_object(key="Box", typeclass=ContribContainer, location=self.room1)

    def test_look_in(self):
        # make sure the object is in the container so we can look at it
        self.obj1.location = self.container
        self.call(CmdContainerLook(), "obj in box", "Obj")
        # move it into a non-container object and look at it there too
        self.obj1.location = self.obj2
        self.call(CmdContainerLook(), "obj in obj2", "Obj")
        # make sure normal looking works too
        self.call(CmdContainerLook(), "obj2", "Obj2")
        self.call(CmdContainerLook(), "", "Room")

    def test_get_and_put(self):
        # get normally
        self.call(CmdContainerGet(), "Obj", "You pick up an Obj.")
        # put in the container
        self.call(CmdPut(), "obj in box", "You put an Obj in a Box.")
        # get from the container
        self.call(CmdContainerGet(), "obj from box", "You get an Obj from a Box.")

    def test_locked_get_put(self):
        # lock container
        self.container.locks.add("get_from:false()")
        # move object to container to try getting
        self.obj1.location = self.container
        self.call(CmdContainerGet(), "obj from box", "You can't get things from that.")
        # move object to character to try putting
        self.obj1.location = self.char1
        self.call(CmdPut(), "obj in box", "You can't put things in that.")

    def test_at_capacity_put(self):
        # set container capacity
        self.container.capacity = 1
        # move object to container to fill capacity
        self.obj2.location = self.container
        # move object to character to try putting
        self.obj1.location = self.char1
        self.call(CmdPut(), "obj in box", "You can't fit anything else in a Box.")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-01-28 18:20
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("help", "0002_auto_20170606_1731")]

    operations = [
        migrations.AlterField(
            model_name="helpentry",
            name="db_entrytext",
            field=models.TextField(
                blank=True, help_text="the main body of help text", verbose_name="help entry"
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_help_category",
            field=models.CharField(
                default="General",
                help_text="organizes help entries in lists",
                max_length=255,
                verbose_name="help category",
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_key",
            field=models.CharField(
                help_text="key to search for", max_length=255, unique=True, verbose_name="help key"
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_lock_storage",
            field=models.TextField(
                blank=True, help_text="normally view:all().", verbose_name="locks"
            ),
        ),
        migrations.AlterField(
            model_name="helpentry",
            name="db_tags",
            field=models.ManyToManyField(
                blank=True,
                help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                to="typeclasses.Tag",
            ),
        ),
    ]



"""

Server plugin services

This plugin module can define user-created services for the Server to
start.

This module must handle all imports and setups required to start a
twisted service (see examples in evennia.server.server). It must also
contain a function start_plugin_services(application). Evennia will
call this function with the main Server application (so your services
can be added to it). The function should not return anything. Plugin
services are started last in the Server startup process.

"""


def start_plugin_services(server):
    """
    This hook is called by Evennia, last in the Server startup process.

    server - a reference to the main server application.
    """
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.9.9 on 2016-09-05 09:02


from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("comms", "0007_msg_db_tags")]

    operations = [
        migrations.AlterModelOptions(name="msg", options={"verbose_name": "Msg"}),
        migrations.RenameField(
            model_name="msg", old_name="db_date_sent", new_name="db_date_created"
        ),
    ]



"""
This allows for running the telnet communication over an encrypted SSL tunnel. To use it, requires a
client supporting Telnet SSL.

The protocol will try to automatically create the private key and certificate on the server side
when starting and will warn if this was not possible. These will appear as files ssl.key and
ssl.cert in mygame/server/.

"""
import os

try:
    from OpenSSL import crypto
    from twisted.internet import ssl as twisted_ssl
except ImportError as error:
    errstr = """
    {err}
    Telnet-SSL requires the PyOpenSSL library and dependencies:

        pip install pyopenssl pycrypto enum pyasn1 service_identity

    Stop and start Evennia again. If no certificate can be generated, you'll
    get a suggestion for a (linux) command to generate this locally.

    """
    raise ImportError(errstr.format(err=error))

from django.conf import settings

from evennia.server.portal.telnet import TelnetProtocol

_GAME_DIR = settings.GAME_DIR

_PRIVATE_KEY_LENGTH = 2048
_PRIVATE_KEY_FILE = os.path.join(_GAME_DIR, "server", "ssl.key")
_PUBLIC_KEY_FILE = os.path.join(_GAME_DIR, "server", "ssl-public.key")
_CERTIFICATE_FILE = os.path.join(_GAME_DIR, "server", "ssl.cert")
_CERTIFICATE_EXPIRE = 365 * 24 * 60 * 60 * 20  # 20 years
_CERTIFICATE_ISSUER = settings.SSL_CERTIFICATE_ISSUER

# messages

NO_AUTOGEN = f"""
Evennia could not auto-generate the SSL private- and public keys ({{err}}).
If this error persists, create them manually (using the tools for your OS). The files
should be placed and named like this:
    {_PRIVATE_KEY_FILE}
    {_PUBLIC_KEY_FILE}
"""

NO_AUTOCERT = """
Evennia's could not auto-generate the SSL certificate ({{err}}).
The private key already exists here:
    {_PRIVATE_KEY_FILE}
If this error persists, create the certificate manually (using the private key and
the tools for your OS). The file should be placed and named like this:
    {_CERTIFICATE_FILE}
"""


class SSLProtocol(TelnetProtocol):
    """
    Communication is the same as telnet, except data transfer
    is done with encryption set up by the portal at start time.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.protocol_key = "telnet/ssl"


def verify_or_create_SSL_key_and_cert(keyfile, certfile):
    """
    Verify or create new key/certificate files.

    Args:
        keyfile (str): Path to ssl.key file.
        certfile (str): Parth to ssl.cert file.

    Notes:
        If files don't already exist, they are created.

    """

    if not (os.path.exists(keyfile) and os.path.exists(certfile)):
        # key/cert does not exist. Create.
        try:
            # generate the keypair
            keypair = crypto.PKey()
            keypair.generate_key(crypto.TYPE_RSA, _PRIVATE_KEY_LENGTH)

            with open(_PRIVATE_KEY_FILE, "wt") as pfile:
                pfile.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, keypair).decode("utf-8"))
                print("Created SSL private key in '{}'.".format(_PRIVATE_KEY_FILE))

            with open(_PUBLIC_KEY_FILE, "wt") as pfile:
                pfile.write(crypto.dump_publickey(crypto.FILETYPE_PEM, keypair).decode("utf-8"))
                print("Created SSL public key in '{}'.".format(_PUBLIC_KEY_FILE))

        except Exception as err:
            print(NO_AUTOGEN.format(err=err))
            return False

        else:
            try:
                # create certificate
                cert = crypto.X509()
                subj = cert.get_subject()
                for key, value in _CERTIFICATE_ISSUER.items():
                    setattr(subj, key, value)
                cert.set_issuer(subj)

                cert.set_serial_number(1000)
                cert.gmtime_adj_notBefore(0)
                cert.gmtime_adj_notAfter(_CERTIFICATE_EXPIRE)
                cert.set_pubkey(keypair)
                cert.sign(keypair, "sha1")

                with open(_CERTIFICATE_FILE, "wt") as cfile:
                    cfile.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"))
                    print("Created SSL certificate in '{}'.".format(_CERTIFICATE_FILE))

            except Exception as err:
                print(NO_AUTOCERT.format(err=err))
                return False

    return True


def getSSLContext():
    """
    This is called by the portal when creating the SSL context
    server-side.

    Returns:
        ssl_context (tuple): A key and certificate that is either
            existing previously or created on the fly.

    """

    if verify_or_create_SSL_key_and_cert(_PRIVATE_KEY_FILE, _CERTIFICATE_FILE):
        return twisted_ssl.DefaultOpenSSLContextFactory(_PRIVATE_KEY_FILE, _CERTIFICATE_FILE)
    else:
        return None



"""
Legacy Comms-commands

Griatch 2021

In Evennia 1.0, the old Channel commands (originally inspired by MUX) were
replaced by the single `channel` command that performs all these function.
That command is still required to talk on channels. This contrib (extracted
from Evennia 0.9.5) reuses the channel-management of the base Channel command
but breaks out its functionality into separate Commands with MUX-familiar names.

- `allcom` - `channel/all` and `channel`
- `addcom` - `channel/alias`, `channel/sub` and `channel/unmute`
- `delcom` - `channel/unalias`, `alias/unsub` and `channel/mute`
- `cboot` - `channel/boot` (`channel/ban` and `/unban` not supported)
- `cwho` - `channel/who`
- `ccreate` - `channel/create`
- `cdestroy` - `channel/destroy`
- `clock` - `channel/lock`
- `cdesc` - `channel/desc`

Installation:

- Import the `CmdSetLegacyComms` cmdset from this module into `mygame/commands/default_cmdsets.py`
- Add it to the CharacterCmdSet's `at_cmdset_creation` method.
- Reload the server.

Example:

```python
# in mygame/commands/default_cmdsets.py

# ..
from evennia.contrib.base_systems.mux_comms_cmds import CmdSetLegacyComms   # <----

class CharacterCmdSet(default_cmds.CharacterCmdSet):
    # ...
    def at_cmdset_creation(self):
        # ...
        self.add(CmdSetLegacyComms)   # <----
```

"""
from django.conf import settings

from evennia.commands.cmdset import CmdSet
from evennia.commands.default.comms import CmdChannel
from evennia.utils import logger

CHANNEL_DEFAULT_TYPECLASS = settings.BASE_CHANNEL_TYPECLASS


class CmdAddCom(CmdChannel):
    """
    Add a channel alias and/or subscribe to a channel

    Usage:
       addcom [alias=] <channel>

    Joins a given channel. If alias is given, this will allow you to
    refer to the channel by this alias rather than the full channel
    name. Subsequent calls of this command can be used to add multiple
    aliases to an already joined channel.
    """

    key = "addcom"
    aliases = ["aliaschan", "chanalias"]
    help_category = "Comms"
    locks = "cmd:not pperm(channel_banned)"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """Implement the command"""

        caller = self.caller
        args = self.args

        if not args:
            self.msg("Usage: addcom [alias =] channelname.")
            return

        if self.rhs:
            # rhs holds the channelname
            channelname = self.rhs
            alias = self.lhs
        else:
            channelname = self.args
            alias = None

        channel = self.search_channel(channelname)
        if not channel:
            return

        string = ""
        if not channel.has_connection(caller):
            # we want to connect as well.
            success, err = self.sub_to_channel(channel)
            if success:
                # if this would have returned True, the account is connected
                self.msg(f"You now listen to the channel {channel.key}")
            else:
                self.msg(f"{channel.key}: You are not allowed to join this channel.")
                return

        if channel.unmute(caller):
            self.msg(f"You unmute channel {channel.key}.")
        else:
            self.msg(f"You are already connected to channel {channel.key}.")

        if alias:
            # create a nick and add it to the caller.
            self.add_alias(channel, alias)
            self.msg(f" You can now refer to the channel {channel} with the alias '{alias}'.")
        else:
            string += " No alias added."
            self.msg(string)


class CmdDelCom(CmdChannel):
    """
    remove a channel alias and/or unsubscribe from channel

    Usage:
       delcom <alias or channel>
       delcom/all <channel>

    If the full channel name is given, unsubscribe from the
    channel. If an alias is given, remove the alias but don't
    unsubscribe. If the 'all' switch is used, remove all aliases
    for that channel.
    """

    key = "delcom"
    aliases = ["delaliaschan", "delchanalias"]
    help_category = "Comms"
    locks = "cmd:not perm(channel_banned)"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """Implementing the command."""

        caller = self.caller

        if not self.args:
            self.msg("Usage: delcom <alias or channel>")
            return
        ostring = self.args.lower().strip()

        channel = self.search_channel(ostring)
        if not channel:
            return

        if not channel.has_connection(caller):
            self.msg("You are not listening to that channel.")
            return

        if ostring == channel.key.lower():
            # an exact channel name - unsubscribe
            delnicks = "all" in self.switches
            # find all nicks linked to this channel and delete them
            if delnicks:
                aliases = self.get_channel_aliases(channel)
                for alias in aliases:
                    self.remove_alias(alias)
            success, err = self.unsub_from_channel(channel)
            if success:
                wipednicks = " Eventual aliases were removed." if delnicks else ""
                self.msg(f"You stop listening to channel '{channel.key}'.{wipednicks}")
            else:
                self.msg(err)
            return
        else:
            # we are removing a channel nick
            self.remove_alias(ostring)
            self.msg(f"Any alias '{ostring}' for channel {channel.key} was cleared.")


class CmdAllCom(CmdChannel):
    """
    perform admin operations on all channels

    Usage:
      allcom [on | off | who | destroy]

    Allows the user to universally turn off or on all channels they are on, as
    well as perform a 'who' for all channels they are on. Destroy deletes all
    channels that you control.

    Without argument, works like comlist.
    """

    key = "allcom"
    aliases = []  # important to not inherit parent's aliases
    locks = "cmd: not pperm(channel_banned)"
    help_category = "Comms"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """Runs the function"""

        caller = self.caller
        args = self.args
        if not args:
            subscribed, available = self.list_channels()
            self.msg("\n|wAvailable channels:\n{table}")
            return
            return

        if args == "on":
            # get names of all channels available to listen to
            # and activate them all
            channels = [
                chan
                for chan in CHANNEL_DEFAULT_TYPECLASS.objects.get_all_channels()
                if chan.access(caller, "listen")
            ]
            for channel in channels:
                self.execute_cmd("addcom %s" % channel.key)
        elif args == "off":
            # get names all subscribed channels and disconnect from them all
            channels = CHANNEL_DEFAULT_TYPECLASS.objects.get_subscriptions(caller)
            for channel in channels:
                self.execute_cmd("delcom %s" % channel.key)
        elif args == "destroy":
            # destroy all channels you control
            channels = [
                chan
                for chan in CHANNEL_DEFAULT_TYPECLASS.objects.get_all_channels()
                if chan.access(caller, "control")
            ]
            for channel in channels:
                self.execute_cmd("cdestroy %s" % channel.key)
        elif args == "who":
            # run a who, listing the subscribers on visible channels.
            string = "\n|CChannel subscriptions|n"
            channels = [
                chan
                for chan in CHANNEL_DEFAULT_TYPECLASS.objects.get_all_channels()
                if chan.access(caller, "listen")
            ]
            if not channels:
                string += "No channels."
            for channel in channels:
                string += "\n|w%s:|n\n %s" % (channel.key, channel.wholist)
            self.msg(string.strip())
        else:
            # wrong input
            self.msg("Usage: allcom on | off | who | clear")


class CmdCdestroy(CmdChannel):
    """
    destroy a channel you created

    Usage:
      cdestroy <channel>

    Destroys a channel that you control.
    """

    key = "cdestroy"
    aliases = []
    help_category = "Comms"
    locks = "cmd: not pperm(channel_banned)"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """Destroy objects cleanly."""

        caller = self.caller

        if not self.args:
            self.msg("Usage: cdestroy <channelname>")
            return

        channel = self.search_channel(self.args)

        if not channel:
            self.msg("Could not find channel %s." % self.args)
            return
        if not channel.access(caller, "control"):
            self.msg("You are not allowed to do that.")
            return
        channel_key = channel.key
        message = f"{channel.key} is being destroyed. Make sure to change your aliases."
        self.destroy_channel(channel, message)
        self.msg("Channel '%s' was destroyed." % channel_key)
        logger.log_sec(
            "Channel Deleted: %s (Caller: %s, IP: %s)."
            % (channel_key, caller, self.session.address)
        )


class CmdCBoot(CmdChannel):
    """
    kick an account from a channel you control

    Usage:
       cboot[/quiet] <channel> = <account> [:reason]

    Switch:
       quiet - don't notify the channel

    Kicks an account or object from a channel you control.

    """

    key = "cboot"
    aliases = []
    switch_options = ("quiet",)
    locks = "cmd: not pperm(channel_banned)"
    help_category = "Comms"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """implement the function"""

        if not self.args or not self.rhs:
            string = "Usage: cboot[/quiet] <channel> = <account> [:reason]"
            self.msg(string)
            return

        channel = self.search_channel(self.lhs)
        if not channel:
            return

        reason = ""
        if ":" in self.rhs:
            target, reason = self.rhs.rsplit(":", 1)
            is_account = target.strip().startswith("*")
            searchstring = target.lstrip("*")
        else:
            is_account = target.strip().startswith("*")
            searchstring = self.rhs.lstrip("*")

        target = self.caller.search(searchstring, account=is_account)
        if not target:
            return
        if reason:
            reason = " (reason: %s)" % reason
        if not channel.access(self.caller, "control"):
            string = "You don't control this channel."
            self.msg(string)
            return

        success, err = self.boot_user(target, quiet="quiet" in self.switches)
        if success:
            self.msg(f"Booted {target.key} from {channel.key}")
            logger.log_sec(
                "Channel Boot: %s (Channel: %s, Reason: %s, Caller: %s, IP: %s)."
                % (self.caller, channel, reason, self.caller, self.session.address)
            )
        else:
            self.msg(err)


class CmdCWho(CmdChannel):
    """
    show who is listening to a channel

    Usage:
      cwho <channel>

    List who is connected to a given channel you have access to.
    """

    key = "cwho"
    aliases = []
    locks = "cmd: not pperm(channel_banned)"
    help_category = "Comms"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """implement function"""

        if not self.args:
            string = "Usage: cwho <channel>"
            self.msg(string)
            return

        channel = self.search_channel(self.lhs)
        if not channel:
            return
        if not channel.access(self.caller, "listen"):
            string = "You can't access this channel."
            self.msg(string)
            return
        string = "\n|CChannel subscriptions|n"
        string += "\n|w%s:|n\n  %s" % (channel.key, channel.wholist)
        self.msg(string.strip())


class CmdChannelCreate(CmdChannel):
    """
    create a new channel

    Usage:
     ccreate <new channel>[;alias;alias...] = description

    Creates a new channel owned by you.
    """

    key = "ccreate"
    aliases = "channelcreate"
    locks = "cmd:not pperm(channel_banned) and pperm(Player)"
    help_category = "Comms"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """Implement the command"""

        if not self.args:
            self.msg("Usage ccreate <channelname>[;alias;alias..] = description")
            return

        description = ""

        if self.rhs:
            description = self.rhs
        lhs = self.lhs
        channame = lhs
        aliases = None
        if ";" in lhs:
            channame, aliases = lhs.split(";", 1)
            aliases = [alias.strip().lower() for alias in aliases.split(";")]

        new_chan, err = self.create_channel(channame, description, aliases=aliases)
        if new_chan:
            self.msg(f"Created channel {new_chan.key} and connected to it.")
        else:
            self.msg(err)


class CmdClock(CmdChannel):
    """
    change channel locks of a channel you control

    Usage:
      clock <channel> [= <lockstring>]

    Changes the lock access restrictions of a channel. If no
    lockstring was given, view the current lock definitions.
    """

    key = "clock"
    aliases = ["clock"]
    locks = "cmd:not pperm(channel_banned) and perm(Admin)"
    help_category = "Comms"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """run the function"""

        if not self.args:
            string = "Usage: clock channel [= lockstring]"
            self.msg(string)
            return

        channel = self.search_channel(self.lhs)
        if not channel:
            return

        if not self.rhs:
            # no =, so just view the current locks
            self.msg(f"Current locks on {channel.key}\n{channel.locks}")
            return
        # we want to add/change a lock.
        if not channel.access(self.caller, "control"):
            string = "You don't control this channel."
            self.msg(string)
            return
        # Try to add the lock
        success, err = self.set_lock(channel, self.rhs)
        if success:
            self.msg(f"Lock(s) applied. Current locks on {channel.key}:\n{channel.locks}")
        else:
            self.msg(err)


class CmdCdesc(CmdChannel):
    """
    describe a channel you control

    Usage:
      cdesc <channel> = <description>

    Changes the description of the channel as shown in
    channel lists.

    """

    key = "cdesc"
    aliases = []
    locks = "cmd:not pperm(channel_banned)"
    help_category = "Comms"

    # this is used by the COMMAND_DEFAULT_CLASS parent
    account_caller = True

    def func(self):
        """Implement command"""

        caller = self.caller

        if not self.rhs:
            self.msg("Usage: cdesc <channel> = <description>")
            return
        channel = self.search_channel(self.lhs)
        if not channel:
            return
        # check permissions
        if not channel.access(caller, "control"):
            self.msg("You cannot admin this channel.")
            return
        self.set_desc(channel, self.rhs)
        self.msg(f"Description of channel '{channel.key}' set to '{self.rhs}'.")


class CmdSetLegacyComms(CmdSet):
    def at_cmdset_createion(self):
        self.add(CmdAddCom())
        self.add(CmdAllCom())
        self.add(CmdDelCom())
        self.add(CmdCdestroy())
        self.add(CmdCBoot())
        self.add(CmdCWho())
        self.add(CmdChannelCreate())
        self.add(CmdClock())
        self.add(CmdCdesc())



"""
Module containing the test cases for the in-game Python system.
"""

from textwrap import dedent

from django.conf import settings
from mock import Mock

from evennia import ScriptDB
from evennia.commands.default.tests import BaseEvenniaCommandTest
from evennia.objects.objects import ExitCommand
from evennia.utils import ansi, utils
from evennia.utils.create import create_object, create_script
from evennia.utils.test_resources import BaseEvenniaTest

from .callbackhandler import CallbackHandler
from .commands import CmdCallback

# Force settings
settings.EVENTS_CALENDAR = "standard"

# Constants
OLD_EVENTS = {}


class TestEventHandler(BaseEvenniaTest):

    """
    Test cases of the event handler to add, edit or delete events.
    """

    def setUp(self):
        """Create the event handler."""
        super().setUp()
        self.handler = create_script(
            "evennia.contrib.base_systems.ingame_python.scripts.EventHandler"
        )

        # Copy old events if necessary
        if OLD_EVENTS:
            self.handler.ndb.events = dict(OLD_EVENTS)

        # Alter typeclasses
        self.char1.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventCharacter"
        )
        self.char2.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventCharacter"
        )
        self.room1.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventRoom"
        )
        self.room2.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventRoom"
        )
        self.exit.swap_typeclass("evennia.contrib.base_systems.ingame_python.typeclasses.EventExit")

    def tearDown(self):
        """Stop the event handler."""
        OLD_EVENTS.clear()
        OLD_EVENTS.update(self.handler.ndb.events)
        self.handler.delete()
        CallbackHandler.script = None
        super().tearDown()

    def test_start(self):
        """Simply make sure the handler runs with proper initial values."""
        self.assertEqual(self.handler.db.callbacks, {})
        self.assertEqual(self.handler.db.to_valid, [])
        self.assertEqual(self.handler.db.locked, [])
        self.assertEqual(self.handler.db.tasks, {})
        self.assertIsNotNone(self.handler.ndb.events)

    def test_add_validation(self):
        """Add a callback while needing validation."""
        author = self.char1
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 40", author=author, valid=False
        )
        callback = self.handler.get_callbacks(self.room1).get("dummy")
        callback = callback[0]
        self.assertIsNotNone(callback)
        self.assertEqual(callback["author"], author)
        self.assertEqual(callback["valid"], False)

        # Since this callback is not valid, it should appear in 'to_valid'
        self.assertIn((self.room1, "dummy", 0), self.handler.db.to_valid)

        # Run this dummy callback (shouldn't do anything)
        self.char1.db.strength = 10
        locals = {"character": self.char1}
        self.assertTrue(self.handler.call(self.room1, "dummy", locals=locals))
        self.assertEqual(self.char1.db.strength, 10)

    def test_edit(self):
        """Test editing a callback."""
        author = self.char1
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 60", author=author, valid=True
        )

        # Edit it right away
        self.handler.edit_callback(
            self.room1, "dummy", 0, "character.db.strength = 65", author=self.char2, valid=True
        )

        # Check that the callback was written
        callback = self.handler.get_callbacks(self.room1).get("dummy")
        callback = callback[0]
        self.assertIsNotNone(callback)
        self.assertEqual(callback["author"], author)
        self.assertEqual(callback["valid"], True)
        self.assertEqual(callback["updated_by"], self.char2)

        # Run this dummy callback
        self.char1.db.strength = 10
        locals = {"character": self.char1}
        self.assertTrue(self.handler.call(self.room1, "dummy", locals=locals))
        self.assertEqual(self.char1.db.strength, 65)

    def test_edit_validation(self):
        """Edit a callback when validation isn't automatic."""
        author = self.char1
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 70", author=author, valid=True
        )

        # Edit it right away
        self.handler.edit_callback(
            self.room1, "dummy", 0, "character.db.strength = 80", author=self.char2, valid=False
        )

        # Run this dummy callback (shouldn't do anything)
        self.char1.db.strength = 10
        locals = {"character": self.char1}
        self.assertTrue(self.handler.call(self.room1, "dummy", locals=locals))
        self.assertEqual(self.char1.db.strength, 10)

    def test_del(self):
        """Try to delete a callback."""
        # Add 3 callbacks
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 5", author=self.char1, valid=True
        )
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 8", author=self.char2, valid=False
        )
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 9", author=self.char1, valid=True
        )

        # Note that the second callback isn't valid
        self.assertIn((self.room1, "dummy", 1), self.handler.db.to_valid)

        # Lock the third callback
        self.handler.db.locked.append((self.room1, "dummy", 2))

        # Delete the first callback
        self.handler.del_callback(self.room1, "dummy", 0)

        # The callback #1 that was to valid should be #0 now
        self.assertIn((self.room1, "dummy", 0), self.handler.db.to_valid)
        self.assertNotIn((self.room1, "dummy", 1), self.handler.db.to_valid)

        # The lock has been updated too
        self.assertIn((self.room1, "dummy", 1), self.handler.db.locked)
        self.assertNotIn((self.room1, "dummy", 2), self.handler.db.locked)

        # Now delete the first (not valid) callback
        self.handler.del_callback(self.room1, "dummy", 0)
        self.assertEqual(self.handler.db.to_valid, [])
        self.assertIn((self.room1, "dummy", 0), self.handler.db.locked)
        self.assertNotIn((self.room1, "dummy", 1), self.handler.db.locked)

        # Call the remaining callback
        self.char1.db.strength = 10
        locals = {"character": self.char1}
        self.assertTrue(self.handler.call(self.room1, "dummy", locals=locals))
        self.assertEqual(self.char1.db.strength, 9)

    def test_accept(self):
        """Accept an callback."""
        # Add 2 callbacks
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 5", author=self.char1, valid=True
        )
        self.handler.add_callback(
            self.room1, "dummy", "character.db.strength = 8", author=self.char2, valid=False
        )

        # Note that the second callback isn't valid
        self.assertIn((self.room1, "dummy", 1), self.handler.db.to_valid)

        # Accept the second callback
        self.handler.accept_callback(self.room1, "dummy", 1)
        callback = self.handler.get_callbacks(self.room1).get("dummy")
        callback = callback[1]
        self.assertIsNotNone(callback)
        self.assertEqual(callback["valid"], True)

        # Call the dummy callback
        self.char1.db.strength = 10
        locals = {"character": self.char1}
        self.assertTrue(self.handler.call(self.room1, "dummy", locals=locals))
        self.assertEqual(self.char1.db.strength, 8)

    def test_call(self):
        """Test to call amore complex callback."""
        self.char1.key = "one"
        self.char2.key = "two"

        # Add an callback
        code = dedent(
            """
            if character.key == "one":
                character.db.health = 50
            else:
                character.db.health = 0
        """.strip(
                "\n"
            )
        )
        self.handler.add_callback(self.room1, "dummy", code, author=self.char1, valid=True)

        # Call the dummy callback
        self.assertTrue(self.handler.call(self.room1, "dummy", locals={"character": self.char1}))
        self.assertEqual(self.char1.db.health, 50)
        self.assertTrue(self.handler.call(self.room1, "dummy", locals={"character": self.char2}))
        self.assertEqual(self.char2.db.health, 0)

    def test_handler(self):
        """Test the object handler."""
        self.assertIsNotNone(self.char1.callbacks)

        # Add an callback
        callback = self.room1.callbacks.add("dummy", "pass", author=self.char1, valid=True)
        self.assertEqual(callback.obj, self.room1)
        self.assertEqual(callback.name, "dummy")
        self.assertEqual(callback.code, "pass")
        self.assertEqual(callback.author, self.char1)
        self.assertEqual(callback.valid, True)
        self.assertIn([callback], list(self.room1.callbacks.all().values()))

        # Edit this very callback
        new = self.room1.callbacks.edit(
            "dummy", 0, "character.db.say = True", author=self.char1, valid=True
        )
        self.assertIn([new], list(self.room1.callbacks.all().values()))
        self.assertNotIn([callback], list(self.room1.callbacks.all().values()))

        # Try to call this callback
        self.assertTrue(self.room1.callbacks.call("dummy", locals={"character": self.char2}))
        self.assertTrue(self.char2.db.say)

        # Delete the callback
        self.room1.callbacks.remove("dummy", 0)
        self.assertEqual(self.room1.callbacks.all(), {})


class TestCmdCallback(BaseEvenniaCommandTest):

    """Test the @callback command."""

    def setUp(self):
        """Create the callback handler."""
        super().setUp()
        self.handler = create_script(
            "evennia.contrib.base_systems.ingame_python.scripts.EventHandler"
        )

        # Copy old events if necessary
        if OLD_EVENTS:
            self.handler.ndb.events = dict(OLD_EVENTS)

        # Alter typeclasses
        self.char1.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventCharacter"
        )
        self.char2.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventCharacter"
        )
        self.room1.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventRoom"
        )
        self.room2.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventRoom"
        )
        self.exit.swap_typeclass("evennia.contrib.base_systems.ingame_python.typeclasses.EventExit")

    def tearDown(self):
        """Stop the callback handler."""
        OLD_EVENTS.clear()
        OLD_EVENTS.update(self.handler.ndb.events)
        self.handler.delete()
        for script in ScriptDB.objects.filter(
            db_typeclass_path="evennia.contrib.base_systems.ingame_python.scripts.TimeEventScript"
        ):
            script.delete()

        CallbackHandler.script = None
        super().tearDown()

    def test_list(self):
        """Test listing callbacks with different rights."""
        table = self.call(CmdCallback(), "out")
        lines = table.splitlines()[3:-1]
        self.assertNotEqual(lines, [])

        # Check that the second column only contains 0 (0) (no callback yet)
        for line in lines:
            cols = line.split("|")
            self.assertIn(cols[2].strip(), ("0 (0)", ""))

        # Add some callback
        self.handler.add_callback(self.exit, "traverse", "pass", author=self.char1, valid=True)

        # Try to obtain more details on a specific callback on exit
        table = self.call(CmdCallback(), "out = traverse")
        lines = table.splitlines()[3:-1]
        self.assertEqual(len(lines), 1)
        line = lines[0]
        cols = line.split("|")
        self.assertIn(cols[1].strip(), ("1", ""))
        self.assertIn(cols[2].strip(), (str(self.char1), ""))
        self.assertIn(cols[-1].strip(), ("Yes", "No", ""))

        # Run the same command with char2
        # char2 shouldn't see the last column (Valid)
        table = self.call(CmdCallback(), "out = traverse", caller=self.char2)
        lines = table.splitlines()[3:-1]
        self.assertEqual(len(lines), 1)
        line = lines[0]
        cols = line.split("|")
        self.assertEqual(cols[1].strip(), "1")
        self.assertNotIn(cols[-1].strip(), ("Yes", "No"))

        # In any case, display the callback
        # The last line should be "pass" (the callback code)
        details = self.call(CmdCallback(), "out = traverse 1")
        self.assertEqual(details.splitlines()[-1], "pass")

    def test_add(self):
        """Test to add an callback."""
        self.call(CmdCallback(), "/add out = traverse")
        editor = self.char1.ndb._eveditor
        self.assertIsNotNone(editor)

        # Edit the callback
        editor.update_buffer(
            dedent(
                """
            if character.key == "one":
                character.msg("You can pass.")
            else:
                character.msg("You can't pass.")
                deny()
        """.strip(
                    "\n"
                )
            )
        )
        editor.save_buffer()
        editor.quit()
        callback = self.exit.callbacks.get("traverse")[0]
        self.assertEqual(callback.author, self.char1)
        self.assertEqual(callback.valid, True)
        self.assertTrue(len(callback.code) > 0)

        # We're going to try the same thing but with char2
        # char2 being a player for our test, the callback won't be validated.
        self.call(CmdCallback(), "/add out = traverse", caller=self.char2)
        editor = self.char2.ndb._eveditor
        self.assertIsNotNone(editor)

        # Edit the callback
        editor.update_buffer(
            dedent(
                """
            character.msg("No way.")
        """.strip(
                    "\n"
                )
            )
        )
        editor.save_buffer()
        editor.quit()
        callback = self.exit.callbacks.get("traverse")[1]
        self.assertEqual(callback.author, self.char2)
        self.assertEqual(callback.valid, False)
        self.assertTrue(len(callback.code) > 0)

    def test_del(self):
        """Add and remove an callback."""
        self.handler.add_callback(self.exit, "traverse", "pass", author=self.char1, valid=True)

        # Try to delete the callback
        # char2 shouldn't be allowed to do so (that's not HIS callback)
        self.call(CmdCallback(), "/del out = traverse 1", caller=self.char2)
        self.assertTrue(len(self.handler.get_callbacks(self.exit).get("traverse", [])) == 1)

        # Now, char1 should be allowed to delete it
        self.call(CmdCallback(), "/del out = traverse 1")
        self.assertTrue(len(self.handler.get_callbacks(self.exit).get("traverse", [])) == 0)

    def test_lock(self):
        """Test the lock of multiple editing."""
        self.call(CmdCallback(), "/add here = time 8:00", caller=self.char2)
        self.assertIsNotNone(self.char2.ndb._eveditor)

        # Now ask char1 to edit
        line = self.call(CmdCallback(), "/edit here = time 1")
        self.assertIsNone(self.char1.ndb._eveditor)

        # Try to delete this callback while char2 is editing it
        line = self.call(CmdCallback(), "/del here = time 1")

    def test_accept(self):
        """Accept an callback."""
        self.call(CmdCallback(), "/add here = time 8:00", caller=self.char2)
        editor = self.char2.ndb._eveditor
        self.assertIsNotNone(editor)

        # Edit the callback
        editor.update_buffer(
            dedent(
                """
            room.msg_contents("It's 8 PM, everybody up!")
        """.strip(
                    "\n"
                )
            )
        )
        editor.save_buffer()
        editor.quit()
        callback = self.room1.callbacks.get("time")[0]
        self.assertEqual(callback.valid, False)

        # chars shouldn't be allowed to the callback
        self.call(CmdCallback(), "/accept here = time 1", caller=self.char2)
        callback = self.room1.callbacks.get("time")[0]
        self.assertEqual(callback.valid, False)

        # char1 will accept the callback
        self.call(CmdCallback(), "/accept here = time 1")
        callback = self.room1.callbacks.get("time")[0]
        self.assertEqual(callback.valid, True)


class TestDefaultCallbacks(BaseEvenniaCommandTest):

    """Test the default callbacks."""

    def setUp(self):
        """Create the callback handler."""
        super().setUp()
        self.handler = create_script(
            "evennia.contrib.base_systems.ingame_python.scripts.EventHandler"
        )

        # Copy old events if necessary
        if OLD_EVENTS:
            self.handler.ndb.events = dict(OLD_EVENTS)

        # Alter typeclasses
        self.char1.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventCharacter"
        )
        self.char2.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventCharacter"
        )
        self.room1.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventRoom"
        )
        self.room2.swap_typeclass(
            "evennia.contrib.base_systems.ingame_python.typeclasses.EventRoom"
        )
        self.exit.swap_typeclass("evennia.contrib.base_systems.ingame_python.typeclasses.EventExit")

    def tearDown(self):
        """Stop the callback handler."""
        OLD_EVENTS.clear()
        OLD_EVENTS.update(self.handler.ndb.events)
        self.handler.delete()
        CallbackHandler.script = None
        super().tearDown()

    def test_exit(self):
        """Test the callbacks of an exit."""
        self.char1.key = "char1"
        code = dedent(
            """
            if character.key == "char1":
                character.msg("You can leave.")
            else:
                character.msg("You cannot leave.")
                deny()
        """.strip(
                "\n"
            )
        )
        # Enforce self.exit.destination since swapping typeclass lose it
        self.exit.destination = self.room2

        # Try the can_traverse callback
        self.handler.add_callback(self.exit, "can_traverse", code, author=self.char1, valid=True)

        # Have char1 move through the exit
        self.call(ExitCommand(), "", "You can leave.", obj=self.exit)
        self.assertIs(self.char1.location, self.room2)

        # Have char2 move through this exit
        self.call(ExitCommand(), "", "You cannot leave.", obj=self.exit, caller=self.char2)
        self.assertIs(self.char2.location, self.room1)

        # Try the traverse callback
        self.handler.del_callback(self.exit, "can_traverse", 0)
        self.handler.add_callback(
            self.exit, "traverse", "character.msg('Fine!')", author=self.char1, valid=True
        )

        # Have char2 move through the exit
        self.call(ExitCommand(), "", obj=self.exit, caller=self.char2)
        self.assertIs(self.char2.location, self.room2)
        self.handler.del_callback(self.exit, "traverse", 0)

        # Move char1 and char2 back
        self.char1.location = self.room1
        self.char2.location = self.room1

        # Test msg_arrive and msg_leave
        code = 'message = "{character} goes out."'
        self.handler.add_callback(self.exit, "msg_leave", code, author=self.char1, valid=True)

        # Have char1 move through the exit
        old_msg = self.char2.msg
        try:
            self.char2.msg = Mock()
            self.call(ExitCommand(), "", obj=self.exit)
            stored_msg = [
                args[0] if args and args[0] else kwargs.get("text", utils.to_str(kwargs))
                for name, args, kwargs in self.char2.msg.mock_calls
            ]
            # Get the first element of a tuple if msg received a tuple instead of a string
            stored_msg = [smsg[0] if isinstance(smsg, tuple) else smsg for smsg in stored_msg]
            returned_msg = ansi.parse_ansi("\n".join(stored_msg), strip_ansi=True)
            self.assertEqual(returned_msg, "char1 goes out.")
        finally:
            self.char2.msg = old_msg

        # Create a return exit
        back = create_object(
            "evennia.objects.objects.DefaultExit",
            key="in",
            location=self.room2,
            destination=self.room1,
        )
        code = 'message = "{character} goes in."'
        self.handler.add_callback(self.exit, "msg_arrive", code, author=self.char1, valid=True)

        # Have char1 move through the exit
        old_msg = self.char2.msg
        try:
            self.char2.msg = Mock()
            self.call(ExitCommand(), "", obj=back)
            stored_msg = [
                args[0] if args and args[0] else kwargs.get("text", utils.to_str(kwargs))
                for name, args, kwargs in self.char2.msg.mock_calls
            ]
            # Get the first element of a tuple if msg received a tuple instead of a string
            stored_msg = [smsg[0] if isinstance(smsg, tuple) else smsg for smsg in stored_msg]
            returned_msg = ansi.parse_ansi("\n".join(stored_msg), strip_ansi=True)
            self.assertEqual(returned_msg, "char1 goes in.")
        finally:
            self.char2.msg = old_msg



"""
Defines a script module with a broken import, to catch the specific error case
in loading global scripts where the module can be parsed but has broken
dependencies.
"""

from evennia import DefaultScript, nonexistent_module


class BrokenScript(DefaultScript):
    pass



"""
These managers define helper methods for accessing the database from
Comm system components.

"""


from django.conf import settings
from django.db.models import Q

from evennia.server import signals
from evennia.typeclasses.managers import TypeclassManager, TypedObjectManager
from evennia.utils import logger
from evennia.utils.utils import class_from_module, dbref, make_iter

_GA = object.__getattribute__
_AccountDB = None
_ObjectDB = None
_ChannelDB = None
_ScriptDB = None
_SESSIONS = None

# error class


class CommError(Exception):
    """
    Raised by comm system, to allow feedback to player when caught.
    """

    pass


#
# helper functions
#


def identify_object(inp):
    """
    Helper function. Identifies if an object is an account or an object;
    return its database model

    Args:
        inp (any): Entity to be idtified.

    Returns:
        identified (tuple): This is a tuple with (`inp`, identifier)
            where `identifier` is one of "account", "object", "channel",
            "string", "dbref" or None.

    """
    if hasattr(inp, "__dbclass__"):
        clsname = inp.__dbclass__.__name__
        if clsname == "AccountDB":
            return inp, "account"
        elif clsname == "ObjectDB":
            return inp, "object"
        elif clsname == "ChannelDB":
            return inp, "channel"
        elif clsname == "ScriptDB":
            return inp, "script"
    if isinstance(inp, str):
        return inp, "string"
    elif dbref(inp):
        return dbref(inp), "dbref"
    else:
        return inp, None


def to_object(inp, objtype="account"):
    """
    Locates the object related to the given accountname or channel key.
    If input was already the correct object, return it.

    Args:
        inp (any): The input object/string
        objtype (str): Either 'account' or 'channel'.

    Returns:
        obj (object): The correct object related to `inp`.

    """
    obj, typ = identify_object(inp)
    if typ == objtype:
        return obj
    if objtype == "account":
        if typ == "object":
            return obj.account
        if typ == "string":
            return _AccountDB.objects.get(user_username__iexact=obj)
        if typ == "dbref":
            return _AccountDB.objects.get(id=obj)
        logger.log_err("%s %s %s %s %s" % (objtype, inp, obj, typ, type(inp)))
        raise CommError()
    elif objtype == "object":
        if typ == "account":
            return obj.obj
        if typ == "string":
            return _ObjectDB.objects.get(db_key__iexact=obj)
        if typ == "dbref":
            return _ObjectDB.objects.get(id=obj)
        logger.log_err("%s %s %s %s %s" % (objtype, inp, obj, typ, type(inp)))
        raise CommError()
    elif objtype == "channel":
        if typ == "string":
            return _ChannelDB.objects.get(db_key__iexact=obj)
        if typ == "dbref":
            return _ChannelDB.objects.get(id=obj)
        logger.log_err("%s %s %s %s %s" % (objtype, inp, obj, typ, type(inp)))
        raise CommError()
    elif objtype == "script":
        if typ == "string":
            return _ScriptDB.objects.get(db_key__iexact=obj)
        if typ == "dbref":
            return _ScriptDB.objects.get(id=obj)
        logger.log_err("%s %s %s %s %s" % (objtype, inp, obj, typ, type(inp)))
        raise CommError()

    # an unknown
    return None


#
# Msg manager
#


class MsgManager(TypedObjectManager):
    """
    This MsgManager implements methods for searching and manipulating
    Messages directly from the database.

    These methods will all return database objects (or QuerySets)
    directly.

    A Message represents one unit of communication, be it over a
    Channel or via some form of in-game mail system. Like an e-mail,
    it always has a sender and can have any number of receivers (some
    of which may be Channels).

    """

    def identify_object(self, inp):
        """
        Wrapper to identify_object if accessing via the manager directly.

        Args:
            inp (any): Entity to be idtified.

        Returns:
            identified (tuple): This is a tuple with (`inp`, identifier)
                where `identifier` is one of "account", "object", "channel",
                "string", "dbref" or None.

        """
        return identify_object(inp)

    def get_message_by_id(self, idnum):
        """
        Retrieve message by its id.

        Args:
            idnum (int or str): The dbref to retrieve.

        Returns:
            message (Msg): The message.

        """
        try:
            return self.get(id=self.dbref(idnum, reqhash=False))
        except Exception:
            return None

    def get_messages_by_sender(self, sender):
        """
        Get all messages sent by one entity - this could be either a
        account or an object

        Args:
            sender (Account or Object): The sender of the message.

        Returns:
            QuerySet: Matching messages.

        Raises:
            CommError: For incorrect sender types.

        """
        obj, typ = identify_object(sender)
        if typ == "account":
            return self.filter(db_sender_accounts=obj).exclude(db_hide_from_accounts=obj)
        elif typ == "object":
            return self.filter(db_sender_objects=obj).exclude(db_hide_from_objects=obj)
        elif typ == "script":
            return self.filter(db_sender_scripts=obj)
        else:
            raise CommError

    def get_messages_by_receiver(self, recipient):
        """
        Get all messages sent to one given recipient.

        Args:
            recipient (Object, Account or Channel): The recipient of the messages to search for.

        Returns:
            Queryset: Matching messages.

        Raises:
            CommError: If the `recipient` is not of a valid type.

        """
        obj, typ = identify_object(recipient)
        if typ == "account":
            return self.filter(db_receivers_accounts=obj).exclude(db_hide_from_accounts=obj)
        elif typ == "object":
            return self.filter(db_receivers_objects=obj).exclude(db_hide_from_objects=obj)
        elif typ == "script":
            return self.filter(db_receivers_scripts=obj)
        else:
            raise CommError

    def search_message(self, sender=None, receiver=None, freetext=None, dbref=None):
        """
        Search the message database for particular messages. At least
        one of the arguments must be given to do a search.

        Args:
            sender (Object, Account or Script, optional): Get messages sent by a particular sender.
            receiver (Object, Account or Channel, optional): Get messages
                received by a certain account,object or channel
            freetext (str): Search for a text string in a message.  NOTE:
                This can potentially be slow, so make sure to supply one of
                the other arguments to limit the search.
            dbref (int): The exact database id of the message. This will override
                    all other search criteria since it's unique and
                    always gives only one match.

        Returns:
            Queryset: Iterable with 0, 1 or more matches.

        """
        # unique msg id
        if dbref:
            return self.objects.filter(id=dbref)

        # We use Q objects to gradually build up the query - this way we only
        # need to do one database lookup at the end rather than gradually
        # refining with multiple filter:s. Django Note: Q objects can be
        # combined with & and | (=AND,OR). ~ negates the queryset

        # filter by sender (we need __pk to avoid an error with empty Q() objects)
        sender, styp = identify_object(sender)
        if sender:
            spk = sender.pk
        if styp == "account":
            sender_restrict = Q(db_sender_accounts__pk=spk) & ~Q(db_hide_from_accounts__pk=spk)
        elif styp == "object":
            sender_restrict = Q(db_sender_objects__pk=spk) & ~Q(db_hide_from_objects__pk=spk)
        elif styp == "script":
            sender_restrict = Q(db_sender_scripts__pk=spk)
        else:
            sender_restrict = Q()
        # filter by receiver
        receiver, rtyp = identify_object(receiver)
        if receiver:
            rpk = receiver.pk
        if rtyp == "account":
            receiver_restrict = Q(db_receivers_accounts__pk=rpk) & ~Q(db_hide_from_accounts__pk=rpk)
        elif rtyp == "object":
            receiver_restrict = Q(db_receivers_objects__pk=rpk) & ~Q(db_hide_from_objects__pk=rpk)
        elif rtyp == "script":
            receiver_restrict = Q(db_receivers_scripts__pk=rpk)
        elif rtyp == "channel":
            raise DeprecationWarning(
                "Msg.objects.search don't accept channel recipients since "
                "Channels no longer accepts Msg objects."
            )
        else:
            receiver_restrict = Q()
        # filter by full text
        if freetext:
            fulltext_restrict = Q(db_header__icontains=freetext) | Q(db_message__icontains=freetext)
        else:
            fulltext_restrict = Q()
        # execute the query
        return self.filter(sender_restrict & receiver_restrict & fulltext_restrict)

    # back-compatibility alias
    message_search = search_message

    def create_message(
        self,
        senderobj,
        message,
        receivers=None,
        locks=None,
        tags=None,
        header=None,
        **kwargs,
    ):
        """
        Create a new communication Msg. Msgs represent a unit of
        database-persistent communication between entites.

        Args:
            senderobj (Object, Account, Script, str or list): The entity (or
                entities) sending the Msg. If a `str`, this is the id-string
                for an external sender type.
            message (str): Text with the message. Eventual headers, titles
                etc should all be included in this text string. Formatting
                will be retained.
            receivers (Object, Account, Script, str or list): An Account/Object to send
                to, or a list of them. If a string, it's an identifier for an external
                receiver.
            locks (str): Lock definition string.
            tags (list): A list of tags or tuples `(tag[,category[,data]])`.
            header (str): Mime-type or other optional information for the message

        Notes:
            The Comm system is created to be very open-ended, so it's fully
            possible to let a message both go several receivers at the same time,
            it's up to the command definitions to limit this as desired.

        """
        if "channels" in kwargs:
            raise DeprecationWarning(
                "create_message() does not accept 'channel' kwarg anymore "
                "- channels no longer accept Msg objects."
            )

        if not message:
            # we don't allow empty messages.
            return None
        new_message = self.model(db_message=message)
        new_message.save()
        for sender in make_iter(senderobj):
            new_message.senders = sender
        new_message.header = header
        for receiver in make_iter(receivers):
            new_message.receivers = receiver
        if locks:
            new_message.locks.add(locks)
        if tags:
            new_message.tags.batch_add(*tags)

        new_message.save()
        return new_message


#
# Channel manager
#


class ChannelDBManager(TypedObjectManager):
    """
    This ChannelManager implements methods for searching and
    manipulating Channels directly from the database.

    These methods will all return database objects (or QuerySets)
    directly.

    A Channel is an in-game venue for communication. It's essentially
    representation of a re-sender: Users sends Messages to the
    Channel, and the Channel re-sends those messages to all users
    subscribed to the Channel.

    """

    def get_all_channels(self):
        """
        Get all channels.

        Returns:
            channels (list): All channels in game.

        """
        return self.all()

    def get_channel(self, channelkey):
        """
        Return the channel object if given its key.
        Also searches its aliases.

        Args:
            channelkey (str): Channel key to search for.

        Returns:
            channel (Channel or None): A channel match.

        """
        dbref = self.dbref(channelkey)
        if dbref:
            try:
                return self.get(id=dbref)
            except self.model.DoesNotExist:
                pass
        results = self.filter(
            Q(db_key__iexact=channelkey)
            | Q(db_tags__db_tagtype__iexact="alias", db_tags__db_key__iexact=channelkey)
        ).distinct()
        return results[0] if results else None

    def get_subscriptions(self, subscriber):
        """
        Return all channels a given entity is subscribed to.

        Args:
            subscriber (Object or Account): The one subscribing.

        Returns:
            subscriptions (list): Channel subscribed to.

        """
        clsname = subscriber.__dbclass__.__name__
        if clsname == "AccountDB":
            return subscriber.account_subscription_set.all()
        if clsname == "ObjectDB":
            return subscriber.object_subscription_set.all()
        return []

    def search_channel(self, ostring, exact=True):
        """
        Search the channel database for a particular channel.

        Args:
            ostring (str): The key or database id of the channel.
            exact (bool, optional): Require an exact (but not
                case sensitive) match.

        Returns:
            Queryset: Iterable with 0, 1 or more matches.

        """
        dbref = self.dbref(ostring)
        if dbref:
            dbref_match = self.search_dbref(dbref)
            if dbref_match:
                return dbref_match

        if exact:
            channels = self.filter(
                Q(db_key__iexact=ostring)
                | Q(db_tags__db_tagtype__iexact="alias", db_tags__db_key__iexact=ostring)
            ).distinct()
        else:
            channels = self.filter(
                Q(db_key__icontains=ostring)
                | Q(db_tags__db_tagtype__iexact="alias", db_tags__db_key__icontains=ostring)
            ).distinct()
        return channels

    def create_channel(
        self,
        key,
        aliases=None,
        desc=None,
        locks=None,
        keep_log=True,
        typeclass=None,
        tags=None,
        attrs=None,
    ):
        """
        Create A communication Channel. A Channel serves as a central hub
        for distributing Msgs to groups of people without specifying the
        receivers explicitly. Instead accounts may 'connect' to the channel
        and follow the flow of messages. By default the channel allows
        access to all old messages, but this can be turned off with the
        keep_log switch.

        Args:
            key (str): This must be unique.

        Keyword Args:
            aliases (list of str): List of alternative (likely shorter) keynames.
            desc (str): A description of the channel, for use in listings.
            locks (str): Lockstring.
            keep_log (bool): Log channel throughput.
            typeclass (str or class): The typeclass of the Channel (not
                often used).
            tags (list): A list of tags or tuples `(tag[,category[,data]])`.
            attrs (list): List of attributes on form `(name, value[,category[,lockstring]])`

        Returns:
            channel (Channel): A newly created channel.

        """
        typeclass = typeclass if typeclass else settings.BASE_CHANNEL_TYPECLASS

        if isinstance(typeclass, str):
            # a path is given. Load the actual typeclass
            typeclass = class_from_module(typeclass, settings.TYPECLASS_PATHS)

        # create new instance
        new_channel = typeclass(db_key=key)

        # store call signature for the signal
        new_channel._createdict = dict(
            key=key,
            aliases=aliases,
            desc=desc,
            locks=locks,
            keep_log=keep_log,
            tags=tags,
            attrs=attrs,
        )

        # this will trigger the save signal which in turn calls the
        # at_first_save hook on the typeclass, where the _createdict can be
        # used.
        new_channel.save()

        signals.SIGNAL_CHANNEL_POST_CREATE.send(sender=new_channel)

        return new_channel

    # back-compatibility alias
    channel_search = search_channel


class ChannelManager(ChannelDBManager, TypeclassManager):
    """
    Wrapper to group the typeclass manager to a consistent name.
    """

    pass
